"""
This module contains core utility functions, as well as stand-alone functions
that are used across the package.
"""

import datetime
import time
from typing import Dict, Iterable, List, Optional, Union, overload, Tuple

import numpy as np
import pandas as pd
import requests
import requests.compat
from packaging import version

from macrosynergy.management.constants import FREQUENCY_MAP


@overload
def get_cid(ticker: str) -> str: ...


@overload
def get_cid(ticker: Iterable[str]) -> List[str]: ...


@overload
def get_xcat(ticker: str) -> str: ...


@overload
def get_xcat(ticker: Iterable[str]) -> List[str]: ...


@overload
def split_ticker(ticker: str) -> str: ...


@overload
def split_ticker(ticker: Iterable[str]) -> List[str]: ...


def split_ticker(ticker: Union[str, Iterable[str]], mode: str) -> Union[str, List[str]]:
    """
    Returns either the cross-sectional identifier (cid) or the category (xcat) from a
    ticker. The function is overloaded to accept either a single ticker or an iterable
    (e.g. list, tuple, pd.Series, np.array) of tickers.

    :param <str> ticker: The ticker to be converted.
    :param <str> mode: The mode to be used. Must be either "cid" or "xcat".

    Returns
    :return <str>: The cross-sectional identifier or category.
    """
    if not isinstance(mode, str):
        raise TypeError("Argument `mode` must be a string.")

    mode: str = mode.lower().strip()
    if mode not in ["cid", "xcat"]:
        raise ValueError("Argument `mode` must be either 'cid' or 'xcat'.")

    if not isinstance(ticker, str):
        if isinstance(ticker, Iterable):
            if len(ticker) == 0:
                raise ValueError("Argument `ticker` must not be empty.")
            return [split_ticker(t, mode) for t in ticker]
        else:
            raise TypeError(
                "Argument `ticker` must be a string or an iterable of strings."
            )

    if "_" not in ticker:
        raise ValueError(
            "Argument `ticker` must be a string"
            " with at least one underscore."
            f" Received '{ticker}' instead."
        )

    cid, xcat = str(ticker).split("_", 1)
    rStr: str = cid if mode == "cid" else xcat
    if len(rStr.strip()) == 0:
        raise ValueError(
            f"Unable to extract {mode} from ticker {ticker}."
            " Please check the ticker."
        )

    return rStr


def get_cid(ticker: Union[str, Iterable[str]]) -> Union[str, List[str]]:
    """
    Returns the cross-sectional identifier (cid) from a ticker.

    :param <str> ticker: The ticker to be converted.

    Returns
    :return <str>: The cross-sectional identifier.
    """
    return split_ticker(ticker, mode="cid")


def get_xcat(ticker: Union[str, Iterable[str]]) -> str:
    """
    Returns the category (xcat) from a ticker.

    :param <str> ticker: The ticker to be converted.

    Returns
    :return <str>: The category.
    """
    return split_ticker(ticker, mode="xcat")


def is_valid_iso_date(date: str) -> bool:
    if not isinstance(date, str):
        raise TypeError("Argument `date` must be a string.")

    try:
        datetime.datetime.strptime(date, "%Y-%m-%d")
        return True
    except ValueError:
        return False


def convert_iso_to_dq(date: str) -> str:
    if is_valid_iso_date(date):
        r = date.replace("-", "")
        assert len(r) == 8, "Date formatting failed"
        return r
    else:
        raise ValueError("Incorrect date format, should be YYYY-MM-DD")


def convert_dq_to_iso(date: str) -> str:
    if len(date) == 8:
        r = datetime.datetime.strptime(date, "%Y%m%d").strftime("%Y-%m-%d")
        assert is_valid_iso_date(r), "Failed to format date"
        return r
    else:
        raise ValueError("Incorrect date format, should be YYYYMMDD")


def _map_to_business_day_frequency(freq: str, valid_freqs: List[str] = None) -> str:
    """
    Maps a frequency string to a business frequency string.

    :param <str> freq: The frequency string to be mapped.
    :param <List[str]> valid_freqs: The valid frequency strings. If None, defaults to
        ["D", "W". "M", "Q", "A"].
    """
    if not isinstance(freq, str):
        raise TypeError("Argument `freq` must be a string.")

    if valid_freqs is not None:
        if (
            (not isinstance(valid_freqs, list))
            or (len(valid_freqs) == 0)
            or (not all(isinstance(x, str) for x in valid_freqs))
        ):
            raise TypeError(
                "Argument `valid_freqs` must be a non-empty list of strings."
            )

    freq = freq.upper()

    if valid_freqs is None:
        valid_freqs = list(FREQUENCY_MAP.keys())
    else:
        # if all valid_freqs are not Frequncy Map keys, raise error - use set to check
        if not set(valid_freqs).issubset(set(FREQUENCY_MAP.keys())):
            raise ValueError(
                f"`valid_freqs` must be a subset of {list(FREQUENCY_MAP.keys())}."
                " See macrosynergy.management.constants.FREQUENCY_MAP for more details."
            )
    if freq in FREQUENCY_MAP.values():
        freq = list(FREQUENCY_MAP.keys())[list(FREQUENCY_MAP.values()).index(freq)]

    if freq not in valid_freqs:
        raise ValueError(
            f"Frequency must be one of {valid_freqs}, but received {freq}."
        )

    return FREQUENCY_MAP[freq]


def form_full_url(url: str, params: Dict = {}) -> str:
    """
    Forms a full URL from a base URL and a dictionary of parameters.
    Useful for logging and debugging.

    :param <str> url: base URL.
    :param <dict> params: dictionary of parameters.

    :return <str>: full URL
    """
    return requests.compat.quote(
        (f"{url}?{requests.compat.urlencode(params)}" if params else url),
        safe="%/:=&?~#+!$,;'@()*[]",
    )


def common_cids(df: pd.DataFrame, xcats: List[str]):
    """
    Returns a list of cross-sectional identifiers (cids) for which the specified
        categories (xcats) are available.

    :param <pd.Dataframe> df: Standardized JPMaQS DataFrame with necessary columns:
        'cid', 'xcat', 'real_date' and 'value'.
    :param <List[str]> xcats: A list with least two categories whose cross-sectional
        identifiers are being considered.

    return <List[str]>: List of cross-sectional identifiers for which all categories in
        `xcats` are available.
    """

    if not isinstance(df, pd.DataFrame):
        raise TypeError("Argument `df` must be a pandas DataFrame.")

    if not isinstance(xcats, list):
        raise TypeError("Argument `xcats` must be a list.")
    elif not all(isinstance(elem, str) for elem in xcats):
        raise TypeError("Argument `xcats` must be a list of strings.")
    elif len(xcats) < 2:
        raise ValueError("Argument `xcats` must contain at least two category tickers.")
    elif not set(xcats).issubset(set(df["xcat"].unique())):
        raise ValueError("All categories in `xcats` must be present in the DataFrame.")

    cid_sets: List[set] = []
    for xc in xcats:
        sc: set = set(df[df["xcat"] == xc]["cid"].unique())
        if sc:
            cid_sets.append(sc)

    ls: List[str] = list(cid_sets[0].intersection(*cid_sets[1:]))
    return sorted(ls)


def generate_random_date(
    start: Optional[Union[str, datetime.datetime, pd.Timestamp]] = "1990-01-01",
    end: Optional[Union[str, datetime.datetime, pd.Timestamp]] = "2020-01-01",
) -> str:
    """
    Generates a random date between two dates.

    :param <str> start: The start date, in the ISO format (YYYY-MM-DD).
    :param <str> end: The end date, in the ISO format (YYYY-MM-DD).

    Returns
    :return <str>: The random date.
    """

    if not isinstance(start, (str, datetime.datetime, pd.Timestamp)):
        raise TypeError(
            "Argument `start` must be a string, datetime.datetime, or pd.Timestamp."
        )
    if not isinstance(end, (str, datetime.datetime, pd.Timestamp)):
        raise TypeError(
            "Argument `end` must be a string, datetime.datetime, or pd.Timestamp."
        )

    start: pd.Timestamp = pd.Timestamp(start)
    end: pd.Timestamp = pd.Timestamp(end)
    if start == end:
        return start.strftime("%Y-%m-%d")
    else:
        return pd.Timestamp(
            np.random.randint(start.value, end.value, dtype=np.int64)
        ).strftime("%Y-%m-%d")


def get_dict_max_depth(d: dict) -> int:
    """
    Returns the maximum depth of a dictionary.

    :param <dict> d: The dictionary to be searched.

    Returns
    :return <int>: The maximum depth of the dictionary.
    """
    return (
        1 + max(map(get_dict_max_depth, d.values()), default=0)
        if isinstance(d, dict)
        else 0
    )


def rec_search_dict(d: dict, key: str, match_substring: bool = False, match_type=None):
    """
    Recursively searches a dictionary for a key and returns the value
    associated with it.

    :param <dict> d: The dictionary to be searched.
    :param <str> key: The key to be searched for.
    :param <bool> match_substring: If True, the function will return
        the value of the first key that contains the substring
        specified by the key parameter. If False, the function will
        return the value of the first key that matches the key
        parameter exactly. Default is False.
    :param <Any> match_type: If not None, the function will look for
        a key that matches the search parameters and has
        the specified type. Default is None.
    :return Any: The value associated with the key, or None if the key
        is not found.
    """
    if not isinstance(d, dict):
        return None

    for k, v in d.items():
        if match_substring:
            if key in k:
                if match_type is None or isinstance(v, match_type):
                    return v
        else:
            if k == key:
                if match_type is None or isinstance(v, match_type):
                    return v

        if isinstance(v, dict):
            result = rec_search_dict(v, key, match_substring, match_type)
            if result is not None:
                return result

    return None


class Timer(object):
    def __init__(self):
        self.t0 = time.perf_counter()

    def __format__(self, format_spec: str):
        if "r" in format_spec:
            return repr(self).__format__(format_spec)
        elif "f" in format_spec:
            return float(self).__format__(format_spec)
        else:
            return str(self).__format__(format_spec)

    def __str__(self) -> str:
        return f"{self.lap():.2f} seconds"

    def __repr__(self) -> str:
        return f"<Time lapsed {str(self):s}>"

    def __float__(self) -> float:
        return self.lap()

    def timer(self) -> Tuple[float, float]:
        x = time.perf_counter()
        return x, x - self.t0

    def lap(self) -> float:
        self.t0, dt = self.timer()
        return dt


def check_package_version(required_version: str):
    from macrosynergy import __version__ as msy_version

    assert version.parse(msy_version) >= version.parse(required_version), (
        f"Current version {msy_version:s} is less than required {required_version:s}"
        " - please upgrade using `pip install macrosynergy --upgrade`"
    )



from .simulate_quantamental_data import (
    make_qdf,
    make_test_df,
    dataframe_generator,
    generate_lines,
    make_qdf_black,
    simulate_ar,
)

from .simulate_vintage_data import VintageData


__all__ = [
    "make_qdf",
    "make_test_df",
    "dataframe_generator",
    "generate_lines",
    "make_qdf_black",
    "simulate_ar",
    "VintageData",
]



"""
Functions used to visualize correlations across categories or cross-sections of
panels.
"""

import itertools
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import scipy.cluster.hierarchy as sch
import seaborn as sns
from matplotlib import pyplot as plt

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import _map_to_business_day_frequency, reduce_df


def view_correlation(
    df: pd.DataFrame,
    xcats: Union[str, List[str]] = None,
    cids: List[str] = None,
    xcats_secondary: Optional[Union[str, List[str]]] = None,
    cids_secondary: Optional[List[str]] = None,
    start: str = "2000-01-01",
    end: str = None,
    val: str = "value",
    freq: str = None,
    cluster: bool = False,
    lags: dict = None,
    lags_secondary: Optional[dict] = None,
    title: str = None,
    size: Tuple[float] = (14, 8),
    max_color: float = None,
    show: bool = True,
    xcat_labels: Optional[Union[List[str], Dict[str, str]]] = None,
    xcat_secondary_labels: Optional[Union[List[str], Dict[str, str]]] = None,
    **kwargs: Any,
):
    """
    Visualize correlation across categories or cross-sections of panels.

    :param <pd.Dataframe> df: standardized JPMaQS DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and at least one column with values of interest.
    :param <List[str]> xcats: extended categories to be correlated. Default is all in the
        DataFrame. If xcats contains only one category the correlation coefficients
        across cross sections are displayed. If xcats contains more than one category,
        the correlation coefficients across categories are displayed. Additionally, the
        order of the xcats received will be mirrored in the correlation matrix.
    :param <List[str]> cids: cross sections to be correlated. Default is all in the
        DataFrame.
    :param <List[str]> xcats_secondary: an optional second set of extended categories.
        If xcats_secondary is provided, correlations will be calculated between the
        categories in xcats and xcats_secondary.
    :param <List[str]> cids_secondary: an optional second list of cross sections. If
        cids_secondary is provided correlations will be calculated and visualized between
        these two sets.
    :param <str> start: earliest date in ISO format. Default is None and earliest date
        in df is used.
    :param <str> end: latest date in ISO format. Default is None and latest date in df
        is used.
    :param <str> val: name of column that contains the values of interest. Default is
        'value'.
    :param <str> freq: frequency option. Per default the correlations are calculated
        based on the native frequency of the datetimes in 'real_date', which is business
        daily. Down-sampling options include weekly ('W'), monthly ('M'), or quarterly
        ('Q') mean.
    :param <bool> cluster: if True the series in the correlation matrix are reordered
        by hierarchical clustering. Default is False.
    :param <dict> lags: optional dictionary of lags applied to respective categories.
        The key will be the category and the value is the lag or lags. If a
        category has multiple lags applied, pass in a list of lag values. The lag factor
        will be appended to the category name in the correlation matrix.
        If xcats_secondary is not none, this parameter will specify lags for the
        categories in xcats.
        N.B.: Lags can include a 0 if the original should also be correlated.
    :param <dict> lags_secondary: optional dictionary of lags applied to the second set of
        categories if xcats_secondary is provided.
    :param <str> title: chart heading. If none is given, a default title is used.
    :param <Tuple[float]> size: two-element tuple setting width/height of figure. Default
        is (14, 8).
    :param <float> max_color: maximum values of positive/negative correlation
        coefficients for color scale. Default is none. If a value is given it applies
        symmetrically to positive and negative values.
    :param <bool> show: if True the figure will be displayed. Default is True.
    :param xcat_labels: optional list or dictionary of labels for xcats.
        A list should be in the same order as xcats, a dictionary should map
        from each xcat to its label.
    :param xcat_secondary_labels: optional list or dictionary of labels for xcats_secondary.
    :param **kwargs: Arbitrary keyword arguments that are passed to seaborn.heatmap.

    N.B:. The function displays the heatmap of a correlation matrix across categories or
    cross-sections (depending on which parameter has received multiple elements).
    """
    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")
    col_names = ["cid", "xcat", "real_date", val]
    df = df[col_names]

    if freq is not None:
        freq = _map_to_business_day_frequency(freq=freq, valid_freqs=["W", "M", "Q"])

    xcats = xcats if isinstance(xcats, list) else [xcats]

    if max_color is not None:
        assert isinstance(max_color, float), "Parameter max_color must be type <float>."

    min_color = None if max_color is None else -max_color
    mask = None
    xlabel = ""
    ylabel = ""

    missing_data_msg = (
        "The provided dataframe does not contain any data for the "
        "specified categories: {xcats}. Please check the data."
    )

    xcat_labels = _parse_xcat_labels(xcats, xcat_labels)

    # If more than one set of xcats or cids have been supplied.
    if xcats_secondary or cids_secondary:
        if xcats_secondary:
            xcat_secondary_labels = _parse_xcat_labels(
                xcats_secondary, xcat_secondary_labels
            )
            xcats_secondary = (
                xcats_secondary
                if isinstance(xcats_secondary, list)
                else [xcats_secondary]
            )
        else:
            xcats_secondary = xcats
            xcat_secondary_labels = xcat_labels

        if not cids_secondary:
            cids_secondary = cids

        df1, xcats, cids = reduce_df(df.copy(), xcats, cids, start, end, out_all=True)
        df2, xcats_secondary, cids_secondary = reduce_df(
            df.copy(), xcats_secondary, cids_secondary, start, end, out_all=True
        )
        for _df, _xc in zip([df1, df2], [xcats, xcats_secondary]):
            if _df.empty:
                raise ValueError(missing_data_msg.format(xcats=_xc))

        s_date = min(df1["real_date"].min(), df2["real_date"].min()).strftime(
            "%Y-%m-%d"
        )
        e_date = max(df1["real_date"].max(), df2["real_date"].max()).strftime(
            "%Y-%m-%d"
        )

        # If only one xcat, we will compute cross sectional correlation.
        if len(xcats) == 1 and len(xcats_secondary) == 1:
            df_w1: pd.DataFrame = _transform_df_for_cross_sectional_corr(
                df=df1, val=val, freq=freq
            )
            df_w2: pd.DataFrame = _transform_df_for_cross_sectional_corr(
                df=df2, val=val, freq=freq
            )

            if title is None:
                title = (
                    f"Cross-sectional correlation of {xcats[0]} and {xcats_secondary[0]} "
                    f"from {s_date} to "
                    f"{e_date}"
                )
            xlabel = f"{xcat_labels[xcats[0]]} cross-sections"
            ylabel = f"{xcat_secondary_labels[xcats_secondary[0]]} cross-sections"

        # If more than one xcat in at least one set, we will compute cross category
        # correlation.
        else:
            df_w1: pd.DataFrame = _transform_df_for_cross_category_corr(
                df=df1, xcats=xcats, val=val, freq=freq, lags=lags
            )
            df_w2: pd.DataFrame = _transform_df_for_cross_category_corr(
                df=df2, xcats=xcats_secondary, val=val, freq=freq, lags=lags_secondary
            )
            df_w1 = df_w1.rename(columns=xcat_labels)
            df_w2 = df_w2.rename(columns=xcat_secondary_labels)
            if title is None:
                title = f"Cross-category correlation from {s_date} to " f"{e_date}"
        corr = (
            pd.concat([df_w1, df_w2], axis=1, keys=["df_w1", "df_w2"])
            .corr()
            .loc["df_w2", "df_w1"]
        )

        if cluster:
            # Since the correlation matrix is not necessarily symmetric, clustering is
            # done in two stages.
            if corr.shape[0] > 1:
                corr = _cluster_correlations(corr=corr, is_symmetric=False)

            if corr.shape[1] > 1:
                corr = _cluster_correlations(corr=corr.T, is_symmetric=False).T

    # If there is only one set of xcats and cids.
    else:
        df, xcats, cids = reduce_df(df, xcats, cids, start, end, out_all=True)
        if df.empty:
            raise ValueError(missing_data_msg.format(xcats=xcats))

        s_date: str = df["real_date"].min().strftime("%Y-%m-%d")
        e_date: str = df["real_date"].max().strftime("%Y-%m-%d")

        if len(xcats) == 1:
            df_w = _transform_df_for_cross_sectional_corr(df=df, val=val, freq=freq)

            if title is None:
                title = (
                    f"Cross-sectional correlation of {xcats[0]} from {s_date} to "
                    f"{e_date}"
                )

        else:
            df_w = _transform_df_for_cross_category_corr(
                df=df, xcats=xcats, val=val, freq=freq, lags=lags
            )

            if title is None:
                title = f"Cross-category correlation from {s_date} to {e_date}"

        df_w = df_w.rename(columns=xcat_labels)
        corr = df_w.corr(method="pearson")

        if cluster:
            corr = _cluster_correlations(corr=corr, is_symmetric=True)

        # Mask for the upper triangle.
        # Return a copy of an array with the elements below the k-th diagonal zeroed. The
        # mask is implemented because correlation coefficients are symmetric.
        mask: np.ndarray = np.triu(np.ones_like(corr, dtype=bool))

    ax: plt.Axes
    fig, ax = plt.subplots(figsize=size)

    with sns.axes_style("white"):
        with sns.axes_style("ticks"):
            sns.heatmap(
                corr,
                mask=mask,
                cmap="vlag_r",
                center=0,
                vmin=min_color,
                vmax=max_color,
                square=False,
                linewidths=0.5,
                cbar_kws={"shrink": 0.5},
                **kwargs,
            )

    ax.set(xlabel=xlabel, ylabel=ylabel)
    ax.set_title(title, fontsize=14)

    if show:
        plt.show()


def _parse_xcat_labels(xcats: List[str], xcat_labels: Union[List[str], Dict[str, str]]):
    """
    Parse xcat labels for correlation plot.

    :param xcats: extended categories to be correlated.
    :param xcat_labels: optional list or dictionary of
        labels for the extended categories.
    """
    labels_dict = {}
    if xcat_labels is not None:
        assert len(xcat_labels) == len(xcats), (
            "The number of labels provided for the extended categories must match the "
            "number of extended categories."
        )
        if isinstance(xcat_labels, list):
            for xcat, xcat_label in zip(xcats, xcat_labels):
                labels_dict[xcat] = xcat_label
        elif isinstance(xcat_labels, dict):
            labels_dict = xcat_labels
        else:
            raise ValueError("The xcats parameter must be a list or a dictionary.")
    else:
        labels_dict = {xcat: xcat for xcat in xcats}
    return labels_dict


def _transform_df_for_cross_sectional_corr(
    df: pd.DataFrame, val: str = "value", freq: str = None
) -> pd.DataFrame:
    """
    Pivots dataframe and down-samples according to the specified frequency so that
    correlation can be calculated between cross-sections.

    :param <pd.Dataframe> df: standardized JPMaQS DataFrame.
    :param <str> val: name of column that contains the values of interest. Default is
        'value'.
    :param <str> freq: Down-sampling frequency option. By default values are not
        down-sampled.

    :return <pd.Dataframe>: The transformed dataframe.
    """
    df_w = df.pivot(index="real_date", columns="cid", values=val)
    if freq is not None:
        df_w = df_w.resample(freq).mean()

    return df_w


def _transform_df_for_cross_category_corr(
    df: pd.DataFrame, xcats: List[str], val: str, freq: str = None, lags: dict = None
) -> pd.DataFrame:
    """
    Pivots dataframe and down-samples according to the specified frequency so that
    correlation can be calculated between extended categories.

    :param <pd.Dataframe> df: standardized JPMaQS DataFrame.
    :param <List[str]> xcats: extended categories to be correlated.
    :param <str> val: name of column that contains the values of interest. Default is
        'value'.
    :param <str> freq: Down-sampling frequency. By default values are not down-sampled.
    :param <dict> lags: optional dictionary of lags applied to respective categories.

    :return <pd.Dataframe>: The transformed dataframe.
    """
    df_w: pd.DataFrame = df.pivot(
        index=("cid", "real_date"), columns="xcat", values=val
    )
    # Down-sample according to the passed frequency.
    if freq is not None:
        df_w = df_w.groupby(
            [pd.Grouper(level="cid"), pd.Grouper(level="real_date", freq=freq)]
        ).mean()

    # Apply the lag mechanism, to the respective categories, after the down-sampling.
    if lags is not None:
        df_w, xcat_tracker = lag_series(df_w=df_w, lags=lags, xcats=xcats)

        # Order the correlation DataFrame to reflect the order of the categories
        # parameter. Will replace the official category name with the lag appended name.
        order = [
            [x] if x not in xcat_tracker.keys() else xcat_tracker[x] for x in xcats
        ]
        order = list(itertools.chain(*order))
    else:
        order = xcats

    df_w = df_w[order]
    return df_w


def lag_series(
    df_w: pd.DataFrame, lags: dict, xcats: List[str]
) -> Tuple[pd.DataFrame, Dict[str, List[str]]]:
    """
    Method used to lag respective categories.

    :param <pd.DataFrame> df_w: multi-index DataFrame where the columns are the
        categories, and the two indices are the cross-sections and real-dates.
    :param <dict> lags: dictionary of lags applied to respective categories.
    :param <List[str]> xcats: extended categories to be correlated.
    """

    lag_type = "The lag data structure must be of type <dict>."
    assert isinstance(lags, dict), lag_type

    lag_xcats = (
        f"The categories referenced in the lagged dictionary must be "
        f"present in the defined DataFrame, {xcats}."
    )
    assert set(lags.keys()).issubset(set(xcats)), lag_xcats

    # Modify the dictionary to adjust for single categories having multiple lags.
    # The respective lags will be held inside a list.
    lag_copy = {}
    xcat_tracker: defaultdict = defaultdict(list)
    for xcat, shift in lags.items():
        if isinstance(shift, int):
            lag_copy[xcat + f"_L{shift}"] = shift
            xcat_tracker[xcat].append(xcat + f"_L{shift}")
        else:
            xcat_temp = [xcat + f"_L{s}" for s in shift]
            # Merge the two dictionaries.
            lag_copy = {**lag_copy, **dict(zip(xcat_temp, shift))}
            xcat_tracker[xcat].extend(xcat_temp)

    df_w_copy = df_w.copy()
    # Handle for multi-index DataFrame. The interior index represents the
    # timestamps.
    for xcat, shift in lag_copy.items():
        category = xcat[:-3]
        clause = isinstance(lags[category], list)
        first_lag = category in df_w.columns

        if clause and not first_lag:
            # Duplicate the column if multiple lags on the same category and the
            # category's first lag has already been implemented. Always access
            # the series from the original DataFrame.
            df_w[xcat] = df_w_copy[category]
        else:
            # Otherwise, modify the name.
            df_w = df_w.rename(columns={category: xcat})
        # Shift the respective column (name will have been adjusted to reflect
        # lag).
        df_w[xcat] = df_w.groupby(level=0)[xcat].shift(shift)

    return df_w, xcat_tracker


def _cluster_correlations(
    corr: pd.DataFrame, is_symmetric: bool = False
) -> pd.DataFrame:
    """
    Rearrange a correlation dataframe so that more similar values are clustered.

    :param <pd.Dataframe> corr: dataframe representing a correlation matrix.
    :param <bool> is_symmetric: if True, rows and columns are rearranged identically.
        If False, only rows are reordered.

    :return <pd.Dataframe>: The sorted correlation dataframe.
    """
    # Get pairwise distances of the dataframe's rows.
    d = sch.distance.pdist(corr)

    # Perform hierarchical / agglomerative clustering. The clustering method used is
    # Farthest Point Algorithm.
    L = sch.linkage(d, method="complete")

    # The second parameter is the distance threshold, t, which will determine the
    # "number" of clusters. If the distance threshold is too small, none of the data
    # points will form a cluster, so n different clusters are returned.
    # If there are any clusters, the categories contained in the cluster will be
    # adjacent.
    ind = sch.fcluster(L, 0.5 * d.max(), "distance")

    indices = [corr.index.tolist()[i] for i in list((np.argsort(ind)))]

    if is_symmetric:
        corr = corr.loc[indices, indices]
    else:
        corr = corr.loc[indices, :]

    return corr


if __name__ == "__main__":
    np.random.seed(0)

    # Un-clustered correlation matrices.

    cids = ["AUD", "CAD", "GBP", "USD", "NZD", "EUR"]
    cids_dmsc = ["CHF", "NOK", "SEK"]
    cids_dmec = ["DEM", "ESP", "FRF", "ITL", "NLG"]
    cids += cids_dmec
    cids += cids_dmsc
    xcats = ["XR", "CRY"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    df_cids.loc["AUD"] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD"] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-11-30", -0.2, 0.5]
    df_cids.loc["USD"] = ["2010-01-01", "2020-12-30", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["EUR"] = ["2002-01-01", "2020-09-30", -0.2, 2]
    df_cids.loc["DEM"] = ["2003-01-01", "2020-09-30", -0.3, 2]
    df_cids.loc["ESP"] = ["2003-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["FRF"] = ["2003-01-01", "2020-09-30", -0.2, 2]
    df_cids.loc["ITL"] = ["2004-01-01", "2020-09-30", -0.2, 0.5]
    df_cids.loc["NLG"] = ["2003-01-01", "2020-12-30", -0.1, 0.5]
    df_cids.loc["CHF"] = ["2003-01-01", "2020-12-30", -0.3, 2.5]
    df_cids.loc["NOK"] = ["2010-01-01", "2020-12-30", -0.1, 0.5]
    df_cids.loc["SEK"] = ["2010-01-01", "2020-09-30", -0.1, 0.5]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR",] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY",] = ["2010-01-01", "2020-10-30", 1, 2, 0.95, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    start = "2012-01-01"
    end = "2020-09-30"

    lag_dict = {"XR": [0, 2, 5]}

    # Clustered correlation matrices. Test hierarchical clustering.
    view_correlation(
        df=dfd,
        xcats=["XR", "CRY"],
        # xcats_secondary=["CRY", "XR"],
        cids=cids,
        cids_secondary=None,
        start=start,
        end=end,
        val="value",
        freq=None,
        cluster=True,
        title="Correlation Matrix",
        size=(14, 8),
        max_color=None,
        lags=None,
        lags_secondary=None,
        annot=True,
        fmt=".2f",
        xcat_labels=["Returns", "Carry"],
        # xcat_secondary_labels={"XR": "Excess returns", "CRY": "Carry"},
    )



"""
Module hosting custom types and meta-classes for use across the package.
"""

from typing import List, Optional
import pandas as pd


class NoneTypeMeta(type):
    """
    MetaClass to support type checks for `None`.
    """

    def __instancecheck__(cls, instance):
        return instance is None or isinstance(instance, type(None))


class NoneType(metaclass=NoneTypeMeta):
    """
    Custom class definition for a NoneType that supports type checks for `None`.
    """

    pass


class QuantamentalDataFrameMeta(type):
    """
    MetaClass to support type checks for `QuantamentalDataFrame`.
    """

    IndexCols: List[str] = ["real_date", "cid", "xcat"]

    def __instancecheck__(cls, instance):
        IDX_COLS = QuantamentalDataFrame.IndexCols
        result: bool = True
        try:
            # the try except offers a safety net in case the instance is not a
            # pd.DataFrame and one of the checks raises an error
            result = result and isinstance(instance, pd.DataFrame)
            result = result and instance.index.name is None
            result = result and not isinstance(instance.columns, pd.MultiIndex)
            result = result and all([col in instance.columns for col in IDX_COLS])
            result = result and len(instance.columns) > len(IDX_COLS)
            result = result and len(instance.columns) == len(set(instance.columns))

            correct_date_type: bool = (
                instance["real_date"].dtype == "datetime64[ns]"
                or isinstance(instance["real_date"].dtype, pd.DatetimeTZDtype)
                or instance.empty
            )
            result = result and correct_date_type

            # # check if the cid col is all str
            # # check if the xcat col is all str
        except:
            result: bool = False
        finally:
            return result


class QuantamentalDataFrame(pd.DataFrame, metaclass=QuantamentalDataFrameMeta):
    """
    ## Type extension of `pd.DataFrame` for Quantamental DataFrames.

    Class definition for a QuantamentalDataFrame that supports type checks for
    `QuantamentalDataFrame`.
    Returns True if the instance is a `pd.DataFrame` with the standard Quantamental
    DataFrame columns ("cid", "xcat", "real_date") and at least one additional column.
    It also checks if the "real_date" column is a datetime type.

    Usage:
    >>> df: pd.DataFrame = make_test_df()
    >>> isinstance(df, QuantamentalDataFrame)
    True
    """

    def __init__(self, df: Optional[pd.DataFrame] = None):
        if df is not None:
            if not (
                isinstance(df, pd.DataFrame) and isinstance(df, QuantamentalDataFrame)
            ):
                raise TypeError("Input must be a QuantamentalDataFrame (pd.DataFrame).")
        super().__init__(df)



'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 0.1.26

version: 0.1.26

full_version: 0.1.26

git_revision: 2d430bf809a99b726a9e691bc4710522d99a9006

release: True

'''
short_version: str = '0.1.26'
version: str = '0.1.26'
full_version: str = '0.1.26'
git_revision: str = '2d430bf809a99b726a9e691bc4710522d99a9006'
release: bool = True
if not release:
    version = full_version



import os
import warnings
from typing import Callable, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from joblib import Parallel, delayed
from sklearn.base import BaseEstimator, clone
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.pipeline import Pipeline
from sklearn.ensemble import VotingRegressor

from collections import defaultdict
from tqdm.auto import tqdm
#from tqdm import tqdm

from macrosynergy.learning import (
    BasePanelSplit,
    ExpandingFrequencyPanelSplit,
    ExpandingKFoldPanelSplit,
    neg_mean_abs_corr,
)
from macrosynergy.management import categories_df, reduce_df, update_df


class BetaEstimator:
    """
    Beta estimation of a panel of contract returns based on regression.

    Estimates betas with seemingly unrelated linear regression (SUR) sequentially with
    expanding panel data windows. Statistical learning determines the model and
    hyperparameter choices sequentially using a given performance metric. Cross-sectional
    betas and out-of-sample "hedged" returns are stored in quantamental dataframes.

    .. note::

      This class is still **experimental** for now: the predictions
      and the API might change without any deprecation cycle.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        xcat: str,
        cids: List[str],
        benchmark_return: str,
    ):
        """
        Initializes BetaEstimator. Takes a quantamental dataframe and creates long
        format dataframes for the contract and benchmark returns.

        :param <pd.DataFrame> df: Daily quantamental dataframe with the following necessary
            columns: 'cid', 'xcat', 'real_date' and 'value'.
        :param <str> xcat: Extended category name for the financial returns to be hedged.
        :param <List[str]> cids: Cross-section identifiers for the financial returns to be hedged.
        :param <str> benchmark_return: Ticker name for the benchmark return to be used in
            the hedging process and beta estimation.
        """
        # Checks
        self._checks_init_params(df, xcat, cids, benchmark_return)

        # Assign class variables
        self.df = df
        self.xcat = xcat
        self.cids = sorted(cids)
        self.benchmark_return = benchmark_return
        self.benchmark_cid, self.benchmark_xcat = self.benchmark_return.split("_", 1)

        # Create pseudo-panel
        dfx = pd.DataFrame(columns=["real_date", "cid", "xcat", "value"])

        for cid in self.cids:
            # Extract cross-section contract returns
            dfa = reduce_df(
                df=self.df,
                xcats=[self.xcat],
                cids=[cid],
            )
            # Extract benchmark returns
            dfb = reduce_df(
                df=self.df,
                xcats=[self.benchmark_xcat],
                cids=[self.benchmark_cid],
            )

            # Combine contract and benchmark returns and rename cross-section identifier
            # in order to match the benchmark return with each cross section in a pseudo
            # panel
            df_cid = pd.concat([dfa, dfb], axis=0)
            df_cid["cid"] = f"{cid}v{self.benchmark_cid}"

            dfx = update_df(dfx, df_cid)

        # Create long format dataframes

        Xy_long = categories_df(
            df=dfx,
            xcats=[self.benchmark_xcat, xcat],
            freq="D",
            xcat_aggs=["sum", "sum"],
            lag=0,
        ).dropna()

        self.X = Xy_long[self.benchmark_xcat]
        self.y = Xy_long[xcat]

        # Create dataframes to store the estimated betas, selected models and chosen
        # hyperparameters
        self.betas = pd.DataFrame(columns=["cid", "real_date", "xcat", "value"])
        self.hedged_returns = pd.DataFrame(
            columns=["cid", "real_date", "xcat", "value"]
        )
        self.chosen_models = pd.DataFrame(
            columns=["real_date", "xcat", "model_type", "hparams", "n_splits"]
        )

    def _checks_init_params(
        self,
        df: pd.DataFrame,
        xcat: str,
        cids: List[str],
        benchmark_return: str,
    ):
        # Dataframe checks
        if not isinstance(df, pd.DataFrame):
            raise TypeError("df must be a pandas DataFrame.")
        if not set(["cid", "xcat", "real_date", "value"]).issubset(df.columns):
            raise ValueError(
                "df must contain columns 'cid', 'xcat', 'real_date' and 'value'."
            )
        if not df["xcat"].nunique() > 1:
            raise ValueError(
                "df must contain at least two xcats. One is required for the contract return panel and another for the benchmark returns."
            )
        # xcat checks
        if not isinstance(xcat, str):
            raise TypeError("xcat must be a string.")
        if xcat not in df["xcat"].unique():
            raise ValueError("xcat must be a valid category in the dataframe.")

        # cids checks
        if not isinstance(cids, list):
            raise TypeError("cids must be a list.")
        if not all(isinstance(cid, str) for cid in cids):
            raise TypeError("All elements in cids must be strings.")
        if not all(cid in df["cid"].unique() for cid in cids):
            raise ValueError(
                "All cids must be valid cross-section identifiers in the dataframe."
            )

        # benchmark return checks
        if not isinstance(benchmark_return, str):
            raise TypeError("benchmark_return must be a string.")
        ticker_list = df["cid"] + "_" + df["xcat"]
        if benchmark_return not in ticker_list.unique():
            raise ValueError(
                "benchmark_return must be a valid ticker in the dataframe."
            )

    def _checks_estimate_beta(
        self,
        beta_xcat,
        hedged_return_xcat,
        inner_splitter,
        scorer,
        models,
        hparam_grid,
        min_cids,
        min_periods,
        est_freq,
        use_variance_correction,
        initial_n_splits,
        threshold_n_periods,
        hparam_type,
        n_iter,
        n_jobs_outer,
        n_jobs_inner,
    ):
        # beta_xcat checks
        if not isinstance(beta_xcat, str):
            raise TypeError("beta_xcat must be a string.")
        if beta_xcat in self.betas["xcat"].unique():
            raise ValueError(
                "beta_xcat already exists in the stored quantamental dataframe for beta "
                "estimates. Please choose a different category name."
            )

        # hedged_return_xcat checks
        if not isinstance(hedged_return_xcat, str):
            raise TypeError("hedged_return_xcat must be a string.")
        if hedged_return_xcat in self.hedged_returns["xcat"].unique():
            raise ValueError(
                "hedged_return_xcat already exists in the stored quantamental dataframe "
                "for hedged returns. Please choose a different category name."
            )

        # inner_splitter checks
        if not isinstance(inner_splitter, BasePanelSplit):
            raise TypeError("inner_splitter must be an instance of BasePanelSplit.")

        # scorer
        # TODO: add to these checks
        if not callable(scorer):
            raise TypeError("scorer must be a callable function.")

        # models grid checks
        if not isinstance(models, dict):
            raise TypeError("models must be a dictionary.")
        if not all(
            isinstance(model, (BaseEstimator, Pipeline)) for model in models.values()
        ):
            raise TypeError(
                "All values in the models dictionary must be scikit-learn compatible models."
            )
        if not all(isinstance(model_name, str) for model_name in models.keys()):
            raise TypeError("All keys in the models dictionary must be strings.")
        for model in models.values():
            if isinstance(model, VotingRegressor):
                for estimator in model.estimators:
                    if not hasattr(estimator[1], "coefs_"):
                        raise ValueError(
                            "All models must be seemingly unrelated linear regressors consistent with "
                            "the scikit-learn API. This means that they must be scikit-learn compatible "
                            "regressors that have a 'coefs_' attribute storing the estimated betas for each "
                            "cross-section, as per the standard imposed by the macrosynergy package. "
                            "Please check that the voting regressor estimators have this attribute."
                        )
            else:
                if not hasattr(model, "coefs_"):
                    raise ValueError(
                        "All models must be seemingly unrelated linear regressors consistent with "
                        "the scikit-learn API. This means that they must be scikit-learn compatible "
                        "regressors that have a 'coefs_' attribute storing the estimated betas for each "
                        "cross-section, as per the standard imposed by the macrosynergy package."
                    )

        # hparam_grid checks
        if not isinstance(hparam_grid, dict):
            raise TypeError("hparam_grid must be a dictionary.")
        if not all(isinstance(model_name, str) for model_name in hparam_grid.keys()):
            raise TypeError("All keys in the hparam_grid dictionary must be strings.")
        if not set(models.keys()) == set(hparam_grid.keys()):
            raise ValueError(
                "The keys in the models and hparam_grid dictionaries must match."
            )
        if not all(isinstance(grid, (dict, list)) for grid in hparam_grid.values()):
            raise TypeError(
                "All values in the hparam_grid dictionary must be dictionaries or lists of dictionaries."
            )
        for grid in hparam_grid.values():
            if isinstance(grid, dict):
                if not all(isinstance(key, str) for key in grid.keys()):
                    raise TypeError(
                        "All keys in the nested dictionaries of hparam_grid must be strings."
                    )
                if not all(isinstance(value, list) for value in grid.values()):
                    raise TypeError(
                        "All values in the nested dictionaries of hparam_grid must be lists."
                    )
            if isinstance(grid, list):
                if not all(isinstance(sub_grid, dict) for sub_grid in grid):
                    raise TypeError(
                        "All elements in the list of hparam_grid must be dictionaries."
                    )
                for sub_grid in grid:
                    if not all(isinstance(key, str) for key in sub_grid.keys()):
                        raise TypeError(
                            "All keys in the nested dictionaries of hparam_grid must be strings."
                        )
                    if not all(isinstance(value, list) for value in sub_grid.values()):
                        raise TypeError(
                            "All values in the nested dictionaries of hparam_grid must be lists."
                        )

        # min_cids checks
        if not isinstance(min_cids, int):
            raise TypeError("min_cids must be an integer.")
        if min_cids < 1:
            raise ValueError("min_cids must be greater than 0.")

        # min_periods checks
        if not isinstance(min_periods, int):
            raise TypeError("min_periods must be an integer.")
        if min_periods < 1:
            raise ValueError("min_periods must be greater than 0.")

        # est_freq checks
        if not isinstance(est_freq, str):
            raise TypeError("est_freq must be a string.")
        if est_freq not in ["D", "W", "M", "Q"]:
            raise ValueError("est_freq must be one of 'D', 'W', 'M' or 'Q'.")

        # use_variance_correction checks
        if not isinstance(use_variance_correction, bool):
            raise TypeError("use_variance_correction must be a boolean.")

        # initial_n_splits checks
        if initial_n_splits is not None:
            if not isinstance(initial_n_splits, int):
                raise TypeError("initial_n_splits must be an integer.")
            if initial_n_splits < 2:
                raise ValueError("initial_n_splits must be greater than 1.")

        # threshold_n_periods checks
        if threshold_n_periods is not None:
            if not isinstance(threshold_n_periods, int):
                raise TypeError("threshold_n_periods must be an integer.")
            if threshold_n_periods < 1:
                raise ValueError("threshold_n_periods must be greater than 0.")

        # hparam_type checks
        if not isinstance(hparam_type, str):
            raise TypeError("hparam_type must be a string.")
        if hparam_type not in ["grid", "prior", "bayes"]:
            raise ValueError("hparam_type must be one of 'grid', 'prior' or 'bayes'.")
        if hparam_type == "bayes":
            raise NotImplementedError(
                "Bayesian hyperparameter search is not yet implemented."
            )

        # n_iter checks
        if n_iter is not None:
            if not isinstance(n_iter, int):
                raise TypeError("n_iter must be an integer.")
            if n_iter < 1:
                raise ValueError("n_iter must be greater than 0.")

        # n_jobs_outer checks
        if not isinstance(n_jobs_outer, int):
            raise TypeError("n_jobs_outer must be an integer.")
        if (n_jobs_outer < -1) or (n_jobs_outer == 0):
            raise ValueError("n_jobs_outer must be greater than zero or equal to -1.")

        # n_jobs_inner checks
        # TODO: raise warning if the product of n_jobs_outer and n_jobs_inner is greater than the number of available cores
        if not isinstance(n_jobs_inner, int):
            raise TypeError("n_jobs_inner must be an integer.")
        if (n_jobs_inner < -1) or (n_jobs_inner == 0):
            raise ValueError("n_jobs_inner must be greater than zero or equal to -1.")

        # Check that the number of jobs isn't excessive
        n_jobs_outer_adjusted = n_jobs_outer if n_jobs_outer != -1 else os.cpu_count()
        n_jobs_inner_adjusted = n_jobs_inner if n_jobs_inner != -1 else os.cpu_count()

        if n_jobs_outer_adjusted * n_jobs_inner_adjusted > os.cpu_count():
            warnings.warn(
                f"n_jobs_outer * n_jobs_inner is greater than the number of available cores. "
                f"Consider reducing the number of jobs to avoid excessive resource usage."
            )

    def estimate_beta(
        self,
        beta_xcat: str,
        hedged_return_xcat: str,
        inner_splitter: BasePanelSplit,
        scorer: Callable,  # Possibly find a better type hint for a scikit-learn scorer
        models: Dict[str, Union[BaseEstimator, Pipeline]],
        hparam_grid: Dict[str, Dict[str, List]],
        min_cids: int = 1,
        min_periods: int = 252,
        est_freq: str = "D",
        use_variance_correction: bool = False,
        initial_n_splits: Optional[
            Union[int, np.int_]
        ] = None,
        threshold_n_periods: Optional[
            Union[int, np.int_]
        ] = None,
        hparam_type: str = "grid",
        n_iter: Optional[int] = 10,
        n_jobs_outer: Optional[int] = -1,
        n_jobs_inner: Optional[int] = 1,
    ):
        """
        Estimates and stores beta coefficients for each cross-section.

        Optimal models and hyperparameters are selected at the end of re-estimation date.
        This includes the frequency of the training dataset. The optimal model is then fit
        to the training set, betas for each cross-section are extracted.  Based in these
        betas hedged returns are calculated up to the next re-estimation date.
        Model and hyperparameter search is based on maximization of a cross-validation
        score from a scikit-learn 'scorer' function.

        :param <str> beta_xcat: Category name for the panel of estimated contract betas.
        :param <str> hedged_return_xcat: Category name for the panel of derived hedged returns.
        :param <BasePanelSplit> inner_splitter: Panel cross-validation splitter for
            the hyperparameter search. It is recommended to set this to an ExpandingKFoldPanelSplit
            splitter.
        :param <Callable> scorer: Scikit-learn scorer function used in both model and
            hyperparameter selection for optimization. For beta estimation, it is recommended
            to use `neg_mean_abs_corr` from the `macrosynergy.learning` submodule.
        :param <dict> models: Dictionary of scikit-learn compatible linear regression models. For beta
            estimation, these models should be seemingly unrelated regressions with a 'coefs_' attribute
            storing estimated betas for each cross-section.
        :param <dict> hparam_grid: Nested dictionary defining the hyperparameters to consider
            for each model.
        :param <int> min_cids: smallest number of cross-sections to be in the initial training set.
            Since market betas are estimated for each cross-section, it is recommended to set this to
            one. This parameter should be used in conjunction with min_periods. Default is 1.
        :param <int> min_periods: minimum requirement for the initial training set in business days.
            This parameter is applied in conjunction with min_cids.
            Default is 1 year (252 days).
        :param <str> est_freq: Frequency forward of each training set for which hedged returns
            are derived. After the hedged returns are determined in each iteration, the training
            panel expands to encapsulate these samples, meaning that this parameter also determines
            re-estimation frequency. This parameter can accept any of the strings "D" (daily),
            "W" (weekly), "M" (monthly) or "Q" (quarterly). Default is "M".
        :param <bool> use_variance_correction: Boolean indicating whether or not to apply a
            correction to cross-validation scores to account for the variation in scores across
            splits. Default is False.
        :param <int> initial_n_splits: Number of splits to be used in cross-validation for the initial
            training set. If None, the number of splits is defined by the inner_splitter. Default is None.
        :param <int> threshold_n_periods: Number of periods, in units of the specified "est_freq"
            to pass before the number of cross-validation splits increases by one. Default is None.
        :param <str> hparam_type: String indicating the type of hyperparameter search. This can be
            `grid`, `prior` or `bayes`. Currently the `bayes` option produces a NotImplementedError.
            Default is "grid".
        :param <int> n_iter: Number of iterations to run for random search. Default is 10.
        :param <int> outer_n_jobs: Number of jobs to run the outer splitter in parallel. Default is -1, which uses
            all available cores.
        :param <int> inner_n_jobs: Number of jobs to run the inner splitter in parallel. Default is 1, which uses
            a single core.
        """
        # Checks
        self._checks_estimate_beta(
            beta_xcat=beta_xcat,
            hedged_return_xcat=hedged_return_xcat,
            inner_splitter=inner_splitter,
            scorer=scorer,
            models=models,
            hparam_grid=hparam_grid,
            min_cids=min_cids,
            min_periods=min_periods,
            est_freq=est_freq,
            use_variance_correction=use_variance_correction,
            initial_n_splits=initial_n_splits,
            threshold_n_periods=threshold_n_periods,
            hparam_type=hparam_type,
            n_iter=n_iter,
            n_jobs_outer=n_jobs_outer,
            n_jobs_inner=n_jobs_inner,
        )

        # (1) Create daily multi-indexed dataframes to store estimated betas and hedged returns
        # This makes indexing easy and the resulting dataframes can be melted to long format later

        min_date = self.X.index.get_level_values(1).min()
        max_date = self.X.index.get_level_values(1).max()
        date_range = pd.bdate_range(start=min_date, end=max_date, freq="B")
        idxs = pd.MultiIndex.from_product(
            [self.cids, date_range], names=["cid", "real_date"]
        )

        stored_betas = pd.DataFrame(
            index=idxs, columns=[beta_xcat], data=np.nan, dtype=np.float64
        )
        stored_hedged_returns = pd.DataFrame(
            index=idxs, columns=[hedged_return_xcat], data=np.nan, dtype=np.float64
        )

        # (2) Set up outer splitter
        outer_splitter = ExpandingFrequencyPanelSplit(
            expansion_freq=est_freq,
            test_freq=est_freq,
            min_cids=min_cids,
            min_periods=min_periods,
        )

        # Create lists to store stored quantities in each iteration
        # This is done to avoid appending to the dataframe in each iteration
        beta_list = []
        hedged_return_list = []
        chosen_models = []

        # (3) Loop through outer splitter, run grid search for optimal model, extract beta
        # estimates, calculate OOS hedged returns and store results
        train_test_splits = list(outer_splitter.split(X=self.X, y=self.y))

        results = Parallel(n_jobs=n_jobs_outer)(
                delayed(self._worker)(
                    train_idx=train_idx,
                    test_idx=test_idx,
                    beta_xcat=beta_xcat,
                    hedged_return_xcat=hedged_return_xcat,
                    inner_splitter=inner_splitter,
                    models=models,
                    scorer=scorer,
                    hparam_grid=hparam_grid,
                    hparam_type=hparam_type,
                    n_iter=n_iter,
                    use_variance_correction=use_variance_correction,
                    initial_n_splits=initial_n_splits,
                    nsplits_add=(
                        np.floor(idx / threshold_n_periods)
                        if initial_n_splits
                        else None
                    ),
                    n_jobs_inner=n_jobs_inner,
                )
                for idx, (train_idx, test_idx) in tqdm(enumerate(train_test_splits),total=len(train_test_splits),)
            )

        for beta_data, hedged_data, model_data in results:
            if beta_data != []:
                beta_list.extend(beta_data)
            if hedged_data != []:
                hedged_return_list.extend(hedged_data)
            chosen_models.append(model_data)

        for cid, real_date, xcat, value in beta_list:
            stored_betas.loc[(cid, real_date), xcat] = value
        for cid, real_date, xcat, value in hedged_return_list:
            stored_hedged_returns.loc[(cid, real_date), xcat] = value

        stored_betas = stored_betas.groupby(level=0).ffill().dropna()
        stored_hedged_returns = stored_hedged_returns.dropna()
        stored_betas_long = pd.melt(
            frame=stored_betas.reset_index(),
            id_vars=["cid", "real_date"],
            var_name="xcat",
            value_name="value",
        )
        stored_hrets_long = pd.melt(
            frame=stored_hedged_returns.reset_index(),
            id_vars=["cid", "real_date"],
            var_name="xcat",
            value_name="value",
        )

        self.betas = pd.concat((self.betas, stored_betas_long), axis=0).astype(
            {
                "cid": "object",
                "real_date": "datetime64[ns]",
                "xcat": "object",
                "value": np.float64,
            }
        )
        self.hedged_returns = pd.concat(
            (self.hedged_returns, stored_hrets_long), axis=0
        ).astype(
            {
                "cid": "object",
                "real_date": "datetime64[ns]",
                "xcat": "object",
                "value": np.float64,
            }
        )

        self.chosen_models = pd.concat(
            (
                self.chosen_models,
                pd.DataFrame(
                    chosen_models,
                    columns=[
                        "real_date",
                        "xcat",
                        "model_type",
                        "hparams",
                        "n_splits",
                    ],
                ),
            )
        )

    def evaluate_hedged_returns(
        self,
        hedged_rets: Optional[Union[str, List[str]]] = None,
        cids: Optional[Union[str, List[str]]] = None,
        correlation_types: Union[str, List[str]] = "pearson",
        title: Optional[str] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        blacklist: Optional[Dict[str, Tuple[pd.Timestamp, pd.Timestamp]]] = None,
        freqs: Optional[Union[str, List[str]]] = "M",
    ):
        """
        Method to determine and display a table of average absolute correlations between 
        the benchmark return and the computed hedged returns within the class instance, over
        all cross-sections in the panel. Additionally, the correlation table displays the
        same results for the unhedged return specified in the class instance for comparison
        purposes.

        The returned dataframe will be multi-indexed by (benchmark return, return, frequency)
        and will contain each computed absolute correlation coefficient on each column.

        :param <Optional[Union[str, List[str]]> hedged_rets: String or list of strings denoting the hedged returns to be
            evaluated. Default is None, which evaluates all hedged returns within the class instance.
        :param <Optional[Union[str, List[str]]> cids: String or list of strings denoting the cross-sections to evaluate.
            Default is None, which evaluates all cross-sections within the class instance.
        :param <Union[str, List[str] correlation_types: String or list of strings denoting the types of correlations
            to calculate. Options are "pearson", "spearman" and "kendall". If None, all three
            are calculated. Default is "pearson".
        :param <Optional[str]> title: Title for the correlation table. If None, the default
            title is "Average absolute correlations between each return and the chosen benchmark". Default is None.
        :param <Optional[str]> start: String in ISO format. Default is None.
        :param <Optional[str]> end: String in ISO format. Default is None.
        :param <Optional[Dict[str, Tuple[pd.Timestamp, pd.Timestamp]]> blacklist: Dictionary of tuples of start and end
            dates to exclude from the evaluation. Default is None.
        :param <Optional[Union[str, List[str]]> freqs: Letters denoting all frequencies 
            at which the series may be sampled. This must be a selection of "D", "W", "M", "Q"
            and "A". Default is "M". Each return series will always be summed over the sample
            period.
        """
        # Checks 
        self._checks_evaluate_hedged_returns(
            correlation_types=correlation_types,
            hedged_rets=hedged_rets,
            cids=cids,
            start=start,
            end=end,
            blacklist=blacklist,
            freqs=freqs,
        )

        # Parameter handling
        if correlation_types is None:
            correlation_types = ["pearson", "spearman", "kendall"]
        elif isinstance(correlation_types, str):
            correlation_types = [correlation_types]
        if hedged_rets is None:
            hedged_rets = list(self.hedged_returns["xcat"].unique())
        elif isinstance(hedged_rets, str):
            hedged_rets = [hedged_rets]
        if cids is None:
            cids = list(self.hedged_returns["cid"].unique())
        elif isinstance(cids, str):
            cids = [cids]
        if isinstance(freqs, str):
            freqs = [freqs]

        # Construct a quantamental dataframe comprising specified hedged returns as well
        # as the unhedged returns and the benchmark return specified in the class instance 
        hedged_df = self.hedged_returns[
            (self.hedged_returns["xcat"].isin(hedged_rets))
            & (self.hedged_returns["cid"].isin(cids))
        ]
        unhedged_df = self.df[
            (self.df["xcat"] == self.xcat) & (self.df["cid"].isin(cids))
        ]
        benchmark_df = self.df[
            (self.df["xcat"] == self.benchmark_xcat) & (self.df["cid"] == self.benchmark_cid)
        ]
        combined_df = pd.concat([hedged_df, unhedged_df], axis=0)

        # Create a pseudo-panel to match contract return cross-sections with a replicated
        # benchmark return. This is multi-indexed by (new cid, real_date). The columns
        # are the named hedged returns, with the final column being the benchmark category.
        dfx = pd.DataFrame(columns=["real_date", "cid", "xcat", "value"])

        for cid in cids:
            # Extract unhedged and hedged returns
            dfa = reduce_df(
                df=combined_df,
                xcats=hedged_rets + [self.xcat],
                cids=[cid],
            )
            # Extract benchmark returns
            dfb = reduce_df(
                df=benchmark_df,
                xcats=[self.benchmark_xcat],
                cids=[self.benchmark_cid],
            )
            # Combine and rename cross-section
            df_cid = pd.concat([dfa, dfb], axis=0)
            df_cid["cid"] = f"{cid}v{self.benchmark_cid}"

            dfx = update_df(dfx, df_cid)

        # Create long format dataframes for each specified frequency
        Xy_long_freq = []
        for freq in freqs:
            Xy_long = categories_df(
                df=dfx, 
                xcats=hedged_rets + [self.xcat, self.benchmark_xcat],
                cids=[f"{cid}v{self.benchmark_cid}" for cid in cids],
                start=start,
                end=end,
                blacklist=blacklist,
                freq=freq,
                xcat_aggs=["sum", "sum"],
            )
            Xy_long_freq.append(Xy_long)

        # For each xcat and frequency, calculate the mean absolute correlations
        # between the benchmark return and the (hedged and unhedged) market returns
        df_rows = []
        for xcat in hedged_rets + [self.xcat]:
            for freq, Xy_long in zip(freqs, Xy_long_freq):
                calculated_correlations = []
                for correlation in correlation_types:
                    calculated_correlations.append(
                        self._get_mean_abs_corrs(
                            xcat=xcat,
                            df=Xy_long,
                            correlation=correlation,
                            cids=cids,
                        )
                    )
                df_rows.append(calculated_correlations)
        # Create underlying dataframe to store the results
        multiindex = pd.MultiIndex.from_product([[self.benchmark_return], hedged_rets + [self.xcat], freqs], names=["benchmark return", "return category", "frequency"])
        corr_df = pd.DataFrame(columns=["|" + correlation + "|" for correlation in correlation_types], index=multiindex, data=df_rows)
        
        return corr_df

    def _worker(
        self,
        train_idx: np.ndarray,
        test_idx: np.ndarray,
        beta_xcat: str,
        hedged_return_xcat: str,
        inner_splitter: BasePanelSplit,
        models: Dict[str, Union[BaseEstimator, Pipeline]],
        scorer: Callable,
        hparam_grid: Dict[str, Dict[str, List]],
        hparam_type: str = "grid",
        n_iter: int = 10,
        use_variance_correction: bool = False,
        initial_n_splits: Optional[Union[int, np.int_]] = None,
        nsplits_add: Optional[Union[int, np.int_]] = None,
        n_jobs_inner: int = 1,
    ):
        # Get training and testing samples
        X_train_i = pd.DataFrame(self.X.iloc[train_idx])
        y_train_i = self.y.iloc[train_idx]
        X_test_i = pd.DataFrame(self.X.iloc[test_idx])
        y_test_i = self.y.iloc[test_idx]
        training_time = X_train_i.index.get_level_values(1).max()

        # Run grid search
        optim_name = None
        optim_model = None
        optim_params = None
        optim_score = -np.inf

        # If nsplits_add is provided, add it to the number of splits
        if initial_n_splits:
            n_splits = initial_n_splits + nsplits_add
            inner_splitter.n_splits = int(n_splits)
        else:
            n_splits = inner_splitter.n_splits

        for model_name, model in models.items():
            if hparam_type == "grid":
                search_object = GridSearchCV(
                    estimator=model,
                    param_grid=hparam_grid[model_name],
                    scoring=scorer,
                    refit=True,
                    cv=inner_splitter,
                    n_jobs=n_jobs_inner,
                )
            elif hparam_type == "random":
                search_object = RandomizedSearchCV(
                    estimator=model,
                    param_distributions=hparam_grid[model_name],
                    n_iter=n_iter,
                    scoring=scorer,
                    refit=True,
                    cv=inner_splitter,
                    n_jobs=n_jobs_inner,
                )

            try:
                search_object.fit(X_train_i, y_train_i)
            except Exception as e:
                warnings.warn(
                    f"Error running a hyperparameter search for {model_name} at {training_time}: {e}",
                    RuntimeWarning,
                )
                continue
            #if not hasattr(search_object, "best_score_"):
            #    # Then the grid search failed completely
            #    warnings.warn(
            #        f"No model was selected at time {training_time}. Hence, no beta can be estimated."
            #    )
            #    # TODO: handle this case properly
            #    continue

            # If a model was selected, extract the score, name, estimator and optimal hyperparameters
            optim_name, optim_model, optim_score, optim_params = (
                self._determine_optimal_model(
                    model_name=model_name,
                    model=model,
                    search_object=search_object,
                    optim_score=optim_score,
                    optim_name=optim_name,
                    optim_model=optim_model,
                    optim_params=optim_params,
                    use_variance_correction=use_variance_correction,
                )
            )

        # Handle case where no model was chosen
        if optim_model is None:
            warnings.warn(
                f"No model was selected at time {training_time}. Hence, no beta can be estimated."
            )
            models_list = [
                training_time,
                beta_xcat,
                "None",
                {},
                inner_splitter.n_splits,
            ]
            beta_list = [
                [cid.split("v")[0], training_time, beta_xcat, np.nan]
                for cid in self.cids
            ]
            hedged_returns = [
                [cid.split("v")[0], training_time] + [hedged_return_xcat] + [np.nan]
                for cid in self.cids
            ]
        else:
            # Get beta estimates for each cross-section
            optim_model.fit(X_train_i, y_train_i)
            if isinstance(optim_model, VotingRegressor):
                estimators = optim_model.estimators_
                coefs_list = [est.coefs_ for est in estimators]
                sum_dict = defaultdict(lambda: [0, 0])

                for coefs in coefs_list:
                    for key, value in coefs.items():
                        sum_dict[key][0] += value
                        sum_dict[key][1] += 1

                betas = {key: sum / count for key, (sum, count) in sum_dict.items()}
            else:
                betas = optim_model.coefs_

            # Get OOS hedged returns
            # This will be a List of lists, with inner lists recording the hedged return
            # for a given cross-section and a given OOS timestamp.
            hedged_returns: List = self._calculate_hedged_returns(
                betas, X_test_i, y_test_i, hedged_return_xcat
            )

            # Compute betas
            beta_list = [
                [cid.split("v")[0], training_time, beta_xcat, beta]
                for cid, beta in betas.items()
            ]

            # Store chosen models and hyperparameters
            models_list = [
                training_time,
                beta_xcat,
                optim_name,
                optim_params,
                inner_splitter.n_splits,
            ]

        return beta_list, hedged_returns, models_list

    def _calculate_hedged_returns(
        self,
        betas,
        X_test_i,
        y_test_i,
        hedged_return_xcat,
    ):
        # TODO: input checks
        betas_series = pd.Series(betas)
        XB = X_test_i.mul(betas_series, level=0, axis=0)
        hedged_returns = y_test_i - XB[self.benchmark_xcat]
        list_hedged_returns = [
            [idx[0].split("v")[0], idx[1]] + [hedged_return_xcat] + [value]
            for idx, value in hedged_returns.items()
        ]

        return list_hedged_returns
    
    def _get_mean_abs_corrs(
        self,
        xcat: str,
        cids: str,
        df: pd.DataFrame,
        correlation: pd.DataFrame,

    ):
        """
        Private helper method to calculate the mean absolute correlation between a column
        'xcat' in a dataframe 'df' and the benchmark return (the last column) across all
        cross-sections in 'cids'. The correlation is calculated using the method specified
        in 'correlation'.
        """
        # Get relevant columns
        df_subset = df[[xcat, self.benchmark_xcat]].dropna()

        # Create inner function to calculate the correlation for a given cross-section
        # This is done so that one can groupby cross-section and apply this function directly

        def calculate_correlation(group):
            return abs(group[xcat].corr(group[self.benchmark_xcat], method=correlation))
       
        # Calculate the mean absolute correlation over all cross sections
        mean_abs_corr = df_subset.groupby("cid").apply(calculate_correlation).mean()

        return mean_abs_corr
    
    def _checks_evaluate_hedged_returns(
        self,
        correlation_types: Union[str, List[str]],
        hedged_rets: Optional[Union[str, List[str]]],
        cids: Optional[Union[str, List[str]]],
        start: Optional[str],
        end: Optional[str],
        blacklist: Optional[Dict[str, Tuple[pd.Timestamp, pd.Timestamp]]],
        freqs: Optional[Union[str, List[str]]],
    ):
        # correlation_types checks
        if (correlation_types is not None) and (not isinstance(correlation_types, (str, list))):
            raise TypeError("correlation_types must be a string or a list of strings.")
        if isinstance(correlation_types, list):
            if not all(isinstance(correlation_type, str) for correlation_type in correlation_types):
                raise TypeError("All elements in correlation_types must be strings.")
            if not all(correlation_type in ["pearson", "spearman", "kendall"] for correlation_type in correlation_types):
                raise ValueError("All elements in correlation_types must be one of 'pearson', 'spearman' or 'kendall'.")
        else:
            if correlation_types not in ["pearson", "spearman", "kendall"]:
                raise ValueError("correlation_types must be one of 'pearson', 'spearman' or 'kendall'.")

        # hedged_rets checks
        if hedged_rets is not None:
            if not isinstance(hedged_rets, (str, list)):
                raise TypeError("hedged_rets must be a string or a list of strings.")
            if isinstance(hedged_rets, list):
                if not all(isinstance(hedged_ret, str) for hedged_ret in hedged_rets):
                    raise TypeError("All elements in hedged_rets must be strings.")
                if not all(hedged_ret in self.hedged_returns["xcat"].unique() for hedged_ret in hedged_rets):
                    raise ValueError(
                        "All hedged_rets must be valid hedged return categories within the class instance."
                    )
            else:
                if hedged_rets not in self.hedged_returns["xcat"].unique():
                    raise ValueError(
                        "hedged_rets must be a valid hedged return category within the class instance."
                    )

        # cids checks
        if cids is not None:
            if not isinstance(cids, (str, list)):
                raise TypeError("cids must be a string or a list of strings.")
            if isinstance(cids, list):
                if not all(isinstance(cid, str) for cid in cids):
                    raise TypeError("All elements in cids must be strings.")
                if not all(cid in self.cids for cid in cids):
                    raise ValueError("All cids must be valid cross-section identifiers within the class instance.")
            else:
                if cids not in self.cids:
                    raise ValueError("cids must be a valid cross-section identifier within the class instance.")

        # start checks
        if start is not None:
            if not isinstance(start, str):
                raise TypeError("start must be a string.")

        # end checks
        if end is not None:
            if not isinstance(end, str):
                raise TypeError("end must be a string.")

        # blacklist checks
        if blacklist is not None:
            if not isinstance(blacklist, dict):
                raise TypeError("The blacklist argument must be a dictionary.")
            for key, value in blacklist.items():
                # check keys are strings
                if not isinstance(key, str):
                    raise TypeError(
                        "The keys of the blacklist argument must be strings."
                    )
                # check values of tuples of length two
                if not isinstance(value, tuple):
                    raise TypeError(
                        "The values of the blacklist argument must be tuples."
                    )
                if len(value) != 2:
                    raise ValueError(
                        "The values of the blacklist argument must be tuples of length "
                        "two."
                    )
                # ensure each of the dates in the dictionary are timestamps
                for date in value:
                    if not isinstance(date, pd.Timestamp):
                        raise TypeError(
                            "The values of the blacklist argument must be tuples of "
                            "pandas Timestamps."
                        )

        # freqs checks
        if freqs is not None:
            if not isinstance(freqs, (str, list)):
                raise TypeError("freqs must be a string or a list of strings.")
            if isinstance(freqs, list):
                if not all(isinstance(freq, str) for freq in freqs):
                    raise TypeError("All elements in freqs must be strings.")
                if not all(freq in ["D", "W", "M", "Q"] for freq in freqs):
                    raise ValueError("All elements in freqs must be one of 'D', 'W', 'M' or 'Q'.")
            else:
                if freqs not in ["D", "W", "M", "Q"]:
                    raise ValueError("freqs must be one of 'D', 'W', 'M' or 'Q'.")
                

    def models_heatmap(
        self,
        beta_xcat: str,
        title: Optional[str] = None,
        cap: Optional[int] = 5,
        figsize: Optional[Tuple[Union[int, float], Union[int, float]]] = (12, 8),
    ):
        """
        Visualizing optimal models used for beta estimation.

        :param <str> beta_xcat: Category name for the panel of estimated contract betas.
        :param <Optional[str]> title: Title of the heatmap. Default is None. This creates
            a figure title of the form "Model Selection Heatmap for {name}".
        :param <Optional[int]> cap: Maximum number of models to display. Default
            (and limit) is 5. The chosen models are the 'cap' most frequently occurring
            in the pipeline.
        :param <Optional[Tuple[Union[int, float], Union[int, float]]]> figsize: Tuple of
            floats or ints denoting the figure size. Default is (12, 8).

        Note:
        This method displays the times at which each model in a learning process
        has been optimal and used for beta estimation, as a binary heatmap.
        """
        # Checks
        self._checks_models_heatmap(
            beta_xcat=beta_xcat, title=title, cap=cap, figsize=figsize
        )

        # Get the chosen models for the specified pipeline to visualise selection.
        chosen_models = self.chosen_models  # TODO: implement self.get_optimal_models()
        chosen_models = chosen_models[chosen_models.xcat == beta_xcat].sort_values(
            by="real_date"
        )
        chosen_models["model_hparam_id"] = chosen_models.apply(
            lambda row: (
                row["model_type"]
                if row["hparams"] == {}
                else f"{row['model_type']}_"
                + "_".join([f"{key}={value}" for key, value in row["hparams"].items()])
            ),
            axis=1,
        )
        chosen_models["real_date"] = chosen_models["real_date"].dt.date
        model_counts = chosen_models.model_hparam_id.value_counts()
        chosen_models = chosen_models[
            chosen_models.model_hparam_id.isin(model_counts.index[:cap])
        ]

        unique_models = chosen_models.model_hparam_id.unique()
        unique_models = sorted(unique_models, key=lambda x: -model_counts[x])
        unique_dates = chosen_models.real_date.unique()

        # Fill in binary matrix denoting the selected model at each time
        binary_matrix = pd.DataFrame(0, index=unique_models, columns=unique_dates)
        for _, row in chosen_models.iterrows():
            model_id = row["model_hparam_id"]
            date = row["real_date"]
            binary_matrix.at[model_id, date] = 1

        # Display the heatmap.
        plt.figure(figsize=figsize)
        if binary_matrix.shape[0] == 1:
            sns.heatmap(binary_matrix, cmap="binary_r", cbar=False)
        else:
            sns.heatmap(binary_matrix, cmap="binary", cbar=False)
        plt.title(title)
        plt.show()

    def _determine_optimal_model(
        self,
        model_name: str,
        model: Union[BaseEstimator, Pipeline],
        search_object: Union[GridSearchCV, RandomizedSearchCV],
        optim_score: float,
        optim_name: str,
        optim_model: Union[BaseEstimator, Pipeline],
        optim_params: Dict,
        use_variance_correction: bool,
    ):
        """
        Private method to determine an optimal model based on cross-validation scores
        at any given estimation time. A given model, with associated model_name, is
        compared to the current optimal model. If the current model has a higher score
        than the optimal model, the current model becomes the optimal model. The new
        optimal score, model name, model and hyperparameters are returned.

        :param <str> model_name: Name of the model being considered.
        :param <Union[BaseEstimator, Pipeline]> model: Model being considered.
        :param <Union[GridSearchCV, RandomizedSearchCV]> search_object: Search object
            containing the results of the hyperparameter search.
        :param <float> optim_score: Current optimal score.
        :param <str> optim_name: Current optimal model name.
        :param <Union[BaseEstimator, Pipeline]> optim_model: Current optimal model.
        :param <Dict> optim_params: Current optimal hyperparameters.
        :param <bool> use_variance_correction: Boolean indicating whether or not to apply a
            correction to cross-validation scores to account for the variation in scores across
            splits.


        """
        if not hasattr(search_object, "cv_results_"):
            return optim_name, optim_model, optim_score, optim_params
        cv_scores = pd.DataFrame(search_object.cv_results_)
        splitscorecols = [
            col
            for col in cv_scores.columns
            if (("split" in col) and ("_test_score" in col))
        ]
        cv_scores = cv_scores[splitscorecols + ["params"]]

        mean_scores = np.nanmean(cv_scores.iloc[:, :-1], axis=1)
        std_scores = np.nanstd(cv_scores.iloc[:, :-1], axis=1)

        if all(np.isnan(mean_scores)):
            return optim_name, optim_model, optim_score, optim_params

        if use_variance_correction:
            # Select model that aims to jointly maximize the mean metric and minimize the
            # standard deviation across splits. Equal weighting is placed on both.
            scaled_mean_scores = (mean_scores - np.nanmin(mean_scores)) / (
                np.nanmax(mean_scores) - np.nanmin(mean_scores)
            )
            scaled_std_scores = (std_scores - np.nanmin(std_scores)) / (
                np.nanmax(std_scores) - np.nanmin(std_scores)
            )
            equalized_scores = (scaled_mean_scores - scaled_std_scores) / 2
            best_index = np.nanargmax(equalized_scores)
            score = equalized_scores[best_index]
        else:
            best_index = np.nanargmax(mean_scores)
            score = mean_scores[best_index]

        if score > optim_score:
            optim_name = model_name
            optim_score = score
            optim_params = cv_scores["params"].values[best_index]
            optim_model = clone(model).set_params(**clone(optim_params, safe=False))

        return optim_name, optim_model, optim_score, optim_params

    def _checks_models_heatmap(
        self,
        beta_xcat: str,
        title: Optional[str] = None,
        cap: Optional[int] = 5,
        figsize: Optional[Tuple[Union[int, float], Union[int, float]]] = (12, 8),
    ):
        if not isinstance(beta_xcat, str):
            raise TypeError("beta_xcat must be a string.")
        if beta_xcat not in self.betas["xcat"].unique():
            raise ValueError(
                "beta_xcat must be a valid category in the stored beta dataframe."
                "Check that estimate_beta has been run with the specified beta_xcat."
            )
        if title is None:
            title = f"Model Selection Heatmap for {beta_xcat}"
        if not isinstance(title, str):
            raise TypeError("title must be a string.")
        if not isinstance(cap, int):
            raise TypeError("cap must be an integer.")
        if cap < 1:
            raise ValueError("cap must be greater than 0.")
        if cap > 20:
            warnings.warn(
                f"The maximum number of models to display is 20. The cap has been set to "
                "20.",
                RuntimeWarning,
            )
            cap = 20
        if not isinstance(figsize, tuple):
            raise TypeError("figsize must be a tuple.")
        if not all(isinstance(value, (int, float)) for value in figsize):
            raise TypeError("All elements in figsize must be integers or floats.")
        if len(figsize) != 2:
            raise ValueError("figsize must be a tuple of length 2.")

    def get_optimal_models(
        self,
        beta_xcat: Optional[Union[str, List]] = None,
    ) -> pd.DataFrame:
        """
        Returns a dataframe of optimal models selected for one or more beta estimation
        processes.

        :param <Optional[Union[str, List]]> beta_xcat: Category name or list of category names
            for the panel of estimated contract betas. If None, information from all
            beta estimation processes held within the class instance is returned. Default is None.

        :return: <pd.DataFrame> A dataframe of optimal models selected for given
            beta estimation processes.
        """
        # Checks
        self._checks_get_optimal_models(beta_xcat=beta_xcat)

        if beta_xcat is None:
            return self.chosen_models
        else:
            return self.chosen_models[self.chosen_models.xcat.isin(beta_xcat)]

    def _checks_get_optimal_models(
        self,
        beta_xcat: Optional[Union[str, List]],
    ):
        if beta_xcat is not None:
            if isinstance(beta_xcat, str):
                beta_xcat = [beta_xcat]
            if not isinstance(beta_xcat, list):
                raise TypeError("beta_xcat must be a string or a list")
            if not all(isinstance(xcat, str) for xcat in beta_xcat):
                raise TypeError(
                    "All elements in beta_xcat, when a list, must be strings."
                )

    def get_hedged_returns(
        self,
        hedged_return_xcat: Optional[Union[str, List]] = None,
    ):
        """
        Returns a dataframe of out-of-sample hedged returns derived from beta estimation processes
        held within the class instance.

        :param <Optional[Union[str, List]]> hedged_return_xcat: Category name or list of category names
            for the panel of derived hedged returns. If None, information from all
            beta estimation processes held within the class instance is returned. Default is None.

        :return: <pd.DataFrame> A dataframe of out-of-sample hedged returns derived from beta estimation
            processes.
        """
        # Checks
        self._checks_get_hedged_returns(hedged_return_xcat=hedged_return_xcat)

        if hedged_return_xcat is None:
            return self.hedged_returns
        else:
            return self.hedged_returns[
                self.hedged_returns.xcat.isin(hedged_return_xcat)
            ]

    def _checks_get_hedged_returns(
        self,
        hedged_return_xcat: Optional[str],
    ):
        if hedged_return_xcat is not None:
            if isinstance(hedged_return_xcat, str):
                hedged_return_xcat = [hedged_return_xcat]
            if not isinstance(hedged_return_xcat, list):
                raise TypeError("hedged_return_xcat must be a string or a list")
            if not all(isinstance(xcat, str) for xcat in hedged_return_xcat):
                raise TypeError(
                    "All elements in hedged_return_xcat, when a list, must be strings."
                )

    def get_betas(
        self,
        beta_xcat: Optional[Union[str, List]] = None,
    ):
        """
        Returns a dataframe of estimated betas derived from beta estimation processes
        held within the class instance.

        :param <Optional[Union[str, List]]> beta_xcat: Category name or list of category names
            for the panel of estimated contract betas. If None, information from all
            beta estimation processes held within the class instance is returned. Default is None.

        :return: <pd.DataFrame> A dataframe of estimated betas derived from beta estimation
            processes.
        """
        # Checks
        self._checks_get_betas(beta_xcat=beta_xcat)

        if beta_xcat is None:
            return self.betas
        else:
            return self.betas[self.betas.xcat.isin(beta_xcat)]

    def _checks_get_betas(
        self,
        beta_xcat: Optional[Union[str, List]],
    ):
        if beta_xcat is not None:
            if isinstance(beta_xcat, str):
                beta_xcat = [beta_xcat]
            if not isinstance(beta_xcat, list):
                raise TypeError("beta_xcat must be a string or a list")
            if not all(isinstance(xcat, str) for xcat in beta_xcat):
                raise TypeError(
                    "All elements in beta_xcat, when a list, must be strings."
                )


if __name__ == "__main__":
    from metrics import neg_mean_abs_corr
    from predictors import LADRegressionSystem, LinearRegressionSystem, CorrelationVolatilitySystem
    from sklearn.ensemble import VotingRegressor
    from macrosynergy.management.simulate import make_qdf

    # Simulate a panel dataset of benchmark and contract returns
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["BENCH_XR", "CONTRACT_XR"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["BENCH_XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CONTRACT_XR"] = ["2001-01-01", "2020-12-31", 0.1, 1, 0, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Initialize the BetaEstimator object
    # Use for the benchmark return: USD_BENCH_XR.
    be = BetaEstimator(
        df=dfd,
        xcat="CONTRACT_XR",
        benchmark_return="USD_BENCH_XR",
        cids=cids,
    )

    models = {
        "LR": LADRegressionSystem(min_xs_samples=21 * 3),
    }
    hparam_grid = {
        "LR": {"fit_intercept": [True, False], "positive": [True, False]}
    }

    scorer = neg_mean_abs_corr

    be.estimate_beta(
        beta_xcat="BETA_NSA",
        hedged_return_xcat="HEDGED_RETURN_NSA",
        inner_splitter=ExpandingKFoldPanelSplit(n_splits = 5),
        scorer=scorer,
        models = models,
        hparam_grid = hparam_grid,
        min_cids=1,
        min_periods=21 * 12,
        est_freq="Q",
        use_variance_correction=False,
        n_jobs_outer=-1,
        n_jobs_inner=1,
    )

    be.models_heatmap(beta_xcat="BETA_NSA")

    evaluation_df = be.evaluate_hedged_returns(
        correlation_types=["pearson", "spearman", "kendall"],
        freqs=["W", "M", "Q"],
    )

    print(evaluation_df)

    models = {
        "VOTE": VotingRegressor(
            [
                ("LR1", LinearRegressionSystem(min_xs_samples=21, data_freq="D")),
                ("LR2", LinearRegressionSystem(min_xs_samples=21, data_freq="W")),
            ]
        )
    }

    hparam_grid = {
        "VOTE" : {
            "LR1__roll": [21, 21 * 3, 21 * 6, 21 * 12],
            "LR2__roll": [3, 6, 9, 12],
        },
    }

    scorer = neg_mean_abs_corr

    be.estimate_beta(
        beta_xcat="BETA_NSA",
        hedged_return_xcat="HEDGED_RETURN_NSA",
        inner_splitter=ExpandingKFoldPanelSplit(n_splits = 5),
        scorer=neg_mean_abs_corr,
        models = models,
        hparam_grid = hparam_grid,
        min_cids=1,
        min_periods=21 * 12,
        est_freq="Q",
        use_variance_correction=False,
        n_jobs_outer=-1,
        n_jobs_inner=1,
    )

    print(be.get_optimal_models())
    print(be.get_betas())
    print(be.get_hedged_returns())
    be.models_heatmap(beta_xcat="BETA_NSA")

    """models = {
        "CORRVOL": CorrelationVolatilitySystem(
            min_xs_samples=21 * 3,
        ),
    }
    hparam_grid = {
        "CORRVOL": [
            #{
            #    "correlation_lookback": [21*12, 21 * 12 * 2, 21 * 12 * 5, 21 * 12 * 10, None],
            #    "correlation_type": ["pearson", "kendall", "spearman"],
            #    "volatility_lookback": [5, 10, 21, 21 * 3, 21 * 6, 21 * 12],
            #    "volatility_window_type": ["exponential", "rolling"],
            #    "data_freq" : ["D"]
            #},
            {
                "correlation_lookback": [4*12, 4 * 12 * 2, 4 * 12 * 5, 4 * 12 * 10, None],
                "correlation_type": ["pearson", "kendall", "spearman"],
                "volatility_lookback": [2, 4, 4 * 3, 4 * 6, 4 * 12],
                "volatility_window_type": ["exponential", "rolling"],
                "data_freq" : ["W"]
            },
        ]
    }
    scorer = neg_mean_abs_corr

    be.estimate_beta(
        beta_xcat="BETA_NSA",
        hedged_return_xcat="HEDGED_RETURN_NSA",
        inner_splitter=ExpandingKFoldPanelSplit(n_splits = 5),
        scorer=neg_mean_abs_corr,
        models = models,
        hparam_grid = hparam_grid,
        min_cids=1,
        min_periods=21 * 3,
        est_freq="Q",
        use_variance_correction=False,
        n_jobs_outer=-1,
        n_jobs_inner=1,
    )

    print(be.get_optimal_models())
    print(be.get_betas())
    print(be.get_hedged_returns())

    be.models_heatmap(beta_xcat="BETA_NSA")
    # Define the models and grids to search over
    models = {
        "LR_ROLL": LinearRegressionSystem(min_xs_samples=21 * 3),
    }
    hparam_grid = {
        "LR_ROLL": [
            {"roll": [21, 21 * 3, 21 * 6, 21 * 12], "data_freq": ["D"]},
            {"roll": [3, 6, 9, 12], "data_freq": ["M"]},
        ]
    }
    scorer = neg_mean_abs_corr

    # Now estimate the betas
    be.estimate_beta(
        beta_xcat="BETA_NSA",
        hedged_return_xcat="HEDGED_RETURN_NSA",
        inner_splitter=ExpandingKFoldPanelSplit(n_splits=5),
        scorer=neg_mean_abs_corr,
        models=models,
        hparam_grid=hparam_grid,
        min_cids=4,
        min_periods=21 * 3,
        est_freq="Q",
        use_variance_correction=True,
        initial_n_splits=5,
        threshold_n_periods=4,
        n_jobs_outer=-1,
        n_jobs_inner=1,
    )

    # Get the results
    print(be.get_optimal_models())
    print(be.get_betas())
    print(be.get_hedged_returns())

    be.models_heatmap(beta_xcat="BETA_NSA")"""



"""
Implementation of `make_relative_category()` function as a module. The function is used
to calculate values for indicators relative to a set of XCATs. This has been developed with sectoral equities returns
in mind but could be extended to other categories where the nature of the data is similar.

It is the twin sister of make_relative_value() as we are aggregating categories for a given CID instead of cids for a
given XCAT.
"""

import pandas as pd
from typing import List, Set

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import reduce_df
import warnings


def _prepare_category_basket(
    df: pd.DataFrame,
    cid: str,
    basket: List[str],
    xcats_avl: List[str],
    complete_set: bool,
):
    """
    Relative cid-specific indicators can be defined over different sets of categories. We will determine the
    respective basket given the available categories for the respective cross-section.

    :param <pd.DataFrame> df: long JPMaQS DataFrame of single category.
    :param <str> cid: target cross-section for the relative value calculation.
    :param <pd.DataFrame> basket: set of categories to be used for the relative value
        benchmark if available.
    :param <List[str] xcats_avl: actual set of categories available for the target cross-section.
    :param <bool> complete_set: if True, the basket is only calculated if all categories, held in the basket,
        are available for that respective category.
    """

    xcats_used: List[str] = sorted(set(basket) & set(xcats_avl))
    xcats_miss: List[str] = list(set(basket) - set(xcats_used))

    # if any missing cids, warn
    if xcats_miss:
        if complete_set:
            xcats_used = []
            err_str: str = (
                f"The cross-section, {cid}, is missing {xcats_miss} which are included "
                f"in the basket {basket}. Therefore, the cross-section will be excluded "
                "from the returned DataFrame."
            )
            print(err_str)
            warnings.warn(err_str, UserWarning)
        else:
            err_str: str = (
                f"The cross-section, {cid}, is missing {xcats_miss} from the requested "
                f"basket. The new basket will be {xcats_used}."
            )
            print(err_str)
            warnings.warn(err_str, UserWarning)

    # Reduce the DataFrame to the specified basket given the available cross-sections.
    dfb = df.loc[df["xcat"].isin(xcats_used)]

    return dfb, xcats_used


def make_relative_category(
    df: pd.DataFrame,
    xcats: List[str] = None,
    cids: List[str] = None,
    start: str = None,
    end: str = None,
    blacklist: dict = None,
    basket: List[str] = None,
    complete_set: bool = False,
    rel_meth: str = "subtract",
    rel_xcats: List[str] = None,
    postfix: str = "RC",
):
    """
    For every given CID, the function returns panel of relative values versus an average of categories.

    :param <pd.DataFrame> df: standardized JPMaQS DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and 'value'.
    :param <List[str]> xcats: all extended categories for which relative values are to be calculated.
        The user must provide the set of xcats to be used in the calculation.
    :param <List[str]> cids: cross-sections for which relative values are calculated.
    :param <str> start: earliest date in ISO format. Default is None and earliest date
        for which the respective category is available is used.
    :param <str> end: latest date in ISO format. Default is None and latest date for
        which the respective category is available is used.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded from
        the output.
    :param <List[str]> basket: categories to be used for the relative value
        benchmark. The default is every categories in the chosen list that is
        available in the DataFrame over the respective time-period.
        However, the basket can be reduced to a valid subset of the available
        categories.
    :param <bool> complete_set: boolean parameter that outlines whether each cid
        is required to have the full set of xcats held by the basket parameter
        for a relative value calculation to occur. If set to True, the cid will be
        excluded from the output if some desired categories are missing.
        Default is False. If False, the mean, for the relative value, will use the subset
        that is available for that category. For instance, if basket = ['AUD', 'CAD',
        'GBP', 'NZD'] but available cids = ['GBP', 'NZD'], the basket will be implicitly
        updated to basket = ['GBP', 'NZD'] for that respective category.
    :param <str> rel_meth: method for calculating relative value. Default is 'subtract'.
        Alternative is 'divide'.
    :param <List[str]> rel_xcats: extended category name of the relative values. Will
        displace the original category names: xcat + postfix. The order should reflect
        the order of the passed categories.
    :param <str> postfix: acronym to be appended to 'xcat' string to give the name for
        relative value category. Only applies if rel_xcats is None. Default is 'RC' relative category

    :return <pd.DataFrame>: standardized DataFrame with the relative values, featuring
        the categories: 'cid', 'xcat', 'real_date' and 'value'.
    """

    col_names = ["cid", "xcat", "real_date", "value"]
    col_error = f"The DataFrame must contain the necessary columns: {col_names}."
    if not set(col_names).issubset(set(df.columns)):
        raise ValueError(col_error)

    operations = {
        "divide": pd.Series.div,
        "subtract": pd.Series.sub,
    }
    if rel_meth not in operations:
        raise ValueError("rel_meth must be 'subtract' or 'divide'")

    if not isinstance(xcats, (list, str)):
        raise TypeError("xcats must be a list of strings or a single string.")

    if isinstance(xcats, str):
        xcats = [xcats]

    if rel_xcats is not None:
        if not (
            isinstance(rel_xcats, list) and all([isinstance(x, str) for x in rel_xcats])
        ):
            raise ValueError("List of strings expected for `rel_xcats`.")

        if len(xcats) != len(rel_xcats):
            raise ValueError(
                "`rel_xcats` must have the same number of elements as `xcats`."
            )

        rel_xcats_dict = dict(zip(xcats, rel_xcats))
    else:
        rel_xcats_dict = {x: x + postfix for x in xcats}

    df = df.loc[:, col_names]
    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")
    # Intersect parameter set to False. Therefore, cross-sections across the categories can vary.
    if basket:
        all_xcats: List[str] = list(set(xcats).union(set(basket)))
    else:
        all_xcats = xcats
    if len(all_xcats) < 1:
        all_xcats = None
    dfx = reduce_df(df, all_xcats, cids, start, end, blacklist, out_all=False)

    if cids is None:
        # All cross-sections available - union across categories.
        cids = list(dfx["cid"].unique())

    if basket is not None:
        # Basket must be a subset of the available xcats
        miss: Set = set(basket) - set(df["xcat"])
        error_basket = f"The category basket elements {miss} are not specified or are not available."
        if not len(miss) == 0:
            raise ValueError(error_basket)
    else:
        # Default basket is all available cross-sections.
        basket = xcats

    available_cids = dfx["cid"].unique()

    if len(xcats) == len(basket) == 1:
        run_error = (
            "Computing the relative value on a single category using a "
            "basket consisting exclusively of the aforementioned category "
            "is an incorrect usage of the function."
        )
        raise ValueError(run_error)

    df_list: List[pd.DataFrame] = []

    # Each cid could have an incomplete basket, we are allowing it to be flexible
    # Each cid processed separately, no contamination!
    for i, cid in enumerate(available_cids):

        df_cid = dfx[dfx["cid"] == cid]
        xcats_avl = df_cid["xcat"].unique()

        df_cid: pd.DataFrame = df_cid[["xcat", "real_date", "value"]]

        dfb, basket = _prepare_category_basket(
            df=df_cid,
            cid=cid,
            basket=basket,
            xcats_avl=xcats_avl,
            complete_set=complete_set,
        )

        if len(basket) > 0:
            bm = dfb.groupby(by="real_date").mean(numeric_only=True).reset_index()
        else:
            continue

        # No need of pivoting, we can operate with groupby()
        dfa = df_cid.copy()
        dfa["count"] = dfa.groupby("real_date")["value"].transform("count")
        dfa = dfa.loc[dfa["count"] > 1]

        # The time-index will be delimited by the respective category.
        dfa = pd.merge(dfa, bm, how="left", on=["real_date"], suffixes=["", "_bm"])
        dfa = dfa.sort_values(by=["xcat", "real_date"])

        dfa["value"] = operations[rel_meth](dfa["value"], dfa["value_bm"], axis=0)

        # cleaning
        df_new = dfa.drop(columns=["value_bm", "count"]).assign(cid=cid)
        df_new["xcat"] = df_new["xcat"].map(rel_xcats_dict)

        if (
            df_new.sort_values(["xcat", "real_date"])[col_names].isna().all().all()
            or (len(df_new) == 0)
            or df_new.empty
        ):
            continue

        df_list.append(df_new.sort_values(["xcat", "real_date"])[col_names])

    return pd.concat(df_list).reset_index(drop=True)


if __name__ == "__main__":
    # Simulate DataFrame.

    cids = ["AUD", "CAD", "GBP", "NZD"]
    xcats = ["XR1", "XR2", "CRY1", "CRY2"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2000-01-01", "2020-12-31", 0.1, 1]
    df_cids.loc["CAD"] = ["2001-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2002-01-01", "2020-11-30", 0, 2]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR1"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["XR2"] = ["2000-01-01", "2020-10-30", 1, 2, 0.95, 1]
    df_xcats.loc["CRY1"] = ["2001-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["CRY2"] = ["2001-01-01", "2020-10-30", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Simulate blacklist
    black = {"AUD": ["2000-01-01", "2003-12-31"], "GBP": ["2018-01-01", "2100-01-01"]}

    # Applications
    dfd_relative_xr = make_relative_category(
        dfd,
        xcats=["XR1", "XR2"],
        cids=None,
        blacklist=None,
        rel_meth="subtract",
        rel_xcats=None,
    )

    dfd_1_black = make_relative_category(
        dfd,
        xcats=["CRY1", "CRY2"],
        cids=None,
        blacklist=black,
        rel_meth="divide",
        rel_xcats=None,
        postfix="_RELRATIO",
    )



"""
Multi PnLs combine multiple "Naive" PnLs with limited signal options and disregarding transaction costs.
"""

from typing import Dict, List, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils.df_utils import update_df
from macrosynergy.pnl import NaivePnL


class MultiPnL:

    def __init__(self):
        self.pnls_df = pd.DataFrame(columns=["real_date", "xcat", "value", "cid"])
        self.single_return_pnls = {}
        self.composite_pnl_xcats = []
        self.xcat_to_ret = {}

    def add_pnl(self, pnl: NaivePnL, pnl_xcats: List[str]):
        """
        Add a NaivePnL object.

        :param pnl: NaivePnL object.
        :param pnl_xcats: List of PnLs to add from the NaivePnL object.
        """
        self._validate_pnl(pnl, pnl_xcats)

        pnl_df = pnl.pnl_df(pnl_xcats)
        pnl_df.loc[:, "xcat"] = pnl_df["xcat"] + "/" + pnl.ret
        # self.pnls_df = pd.concat([self.pnls_df, pnl_df], axis=0, ignore_index=True)
        self.pnls_df = update_df(self.pnls_df, pnl_df)
        for xcat in pnl_df.xcat.unique():
            self.single_return_pnls[xcat] = pnl

        for xcat in pnl_xcats:
            if xcat not in self.xcat_to_ret:
                self.xcat_to_ret[xcat] = {pnl.ret}
            else:
                self.xcat_to_ret[xcat].add(pnl.ret)
        pass

    def combine_pnls(
        self,
        pnl_xcats: List[str],
        composite_pnl_xcat: str,
        weights: Optional[Dict[str, float]] = None,
    ) -> pd.DataFrame:
        """
        Combine PnLs with optional weighting.

        :param pnl_xcats: List of PnLs to combine. Must be in the format 'xcat/return' and
            added using `add_pnl()`.
        :param composite_pnl_xcat: xcat for the combined PnL.
        :param weights: Weights for each PnL, by default None. Must be in the format
            {'xcat': weight} or {'xcat/return': weight}.
        """
        self._check_pnls_added(min_pnls=2)
        for i, pnl_xcat in enumerate(pnl_xcats):
            pnl_xcats[i] = self._infer_return_by_xcat(pnl_xcat)

        # Default weights
        if weights is None:
            weights = {pnl_name: 1 for pnl_name in pnl_xcats}
        else:
            weights = {self._infer_return_by_xcat(k): v for k, v in weights.items()}
        weights = self._normalize_weights(weights)

        multiasset_df = []
        for pnl_xcat in pnl_xcats:
            single_asset_df = self.pnls_df[self.pnls_df["xcat"] == pnl_xcat].assign(
                asset=pnl_xcat
            )
            multiasset_df.append(single_asset_df)

        multiasset_df = pd.concat(multiasset_df, axis=0, ignore_index=True)

        raw_pnls = multiasset_df.set_index(["real_date", "xcat"])["value"].unstack()

        # Default weights for each strategy
        start_weights = pd.DataFrame(
            {asset_name: weights[asset_name] for asset_name in raw_pnls.columns},
            index=raw_pnls.index,
        )

        # Daily change in portfolio weights due to previous returns since the last rebalancing
        weights_change = (
            (1 + raw_pnls / 100).groupby(pd.Grouper(freq="M")).cumprod()
        )  # in decimals, not percentage, gross amount
        weights_change = (
            weights_change.groupby(pd.Grouper(freq="M"))
            .shift(periods=1)
            .fillna(value=1)
        )

        # Dynamic weights
        final_weights = start_weights * weights_change
        final_weights = final_weights.div(final_weights.sum(axis=1), axis=0)

        # final calculation
        multiasset_rets = (final_weights * raw_pnls).sum(axis=1)

        multiasset_rets.name = composite_pnl_xcat

        multi_asset_pnl = multiasset_rets.reset_index().melt(
            id_vars=["real_date"], var_name="xcat", value_name="value"
        )
        multi_asset_pnl = multi_asset_pnl.sort_values(by=["xcat", "real_date"])
        multi_asset_pnl["cid"] = "ALL"

        self.pnls_df = update_df(self.pnls_df, multi_asset_pnl).sort_values(
            by=["xcat", "real_date"]
        )
        self.composite_pnl_xcats.append(composite_pnl_xcat)

    def plot_pnls(
        self,
        pnl_xcats: List[str] = None,
        title: str = None,
        title_fontsize: int = 14,
        xcat_labels: Union[List[str], dict] = None,
    ):
        """
        Creates a plot of PnLs

        :param pnl_xcats: List of PnLs to plot. If None, all PnLs are plotted.
            Must be in the format 'xcat', or 'xcat/return_xcat'.
        :param title: Title of the plot.
        """
        self._check_pnls_added()

        if pnl_xcats is None:
            pnl_df = self.pnls_df
            pnl_xcats = self.pnl_xcats()
        else:
            for i, pnl_xcat in enumerate(pnl_xcats):
                pnl_xcats[i] = self._infer_return_by_xcat(pnl_xcat)
            pnl_df = self.pnls_df[self.pnls_df["xcat"].isin(pnl_xcats)].copy()

        if xcat_labels is not None:

            xcat_labels = self._check_xcat_labels(pnl_xcats, xcat_labels)
            pnl_df["xcat"] = pnl_df["xcat"].map(xcat_labels)

        pnl_df.loc[:, "cumulative pnl"] = pnl_df.groupby("xcat")["value"].cumsum()

        sns.lineplot(data=pnl_df, x="real_date", y="cumulative pnl", hue=("xcat"))
        plt.title(title, fontsize=title_fontsize)
        plt.xlabel(None)
        plt.ylabel("% risk capital, no compounding")
        plt.legend(title="PnL Category(s)")
        plt.show()
        pnl_df.drop(columns="cumulative pnl", inplace=True)

    def evaluate_pnls(self, pnl_xcats: List[str] = None) -> pd.DataFrame:
        """
        Evaluate individual and composite PnLs.

        :param pnl_xcats: List of PnLs to evaluate. If None, all PnLs are evaluated.
            Must be in the format 'xcat', or 'xcat/return_xcat'.
        """
        self._check_pnls_added()
        if pnl_xcats is None:
            pnl_xcats = self.pnl_xcats()
        else:
            for i, pnl_xcat in enumerate(pnl_xcats):
                pnl_xcats[i] = self._infer_return_by_xcat(pnl_xcat)
        pnl_evals = []
        for pnl_xcat in pnl_xcats:
            if pnl_xcat in self.composite_pnl_xcats:
                eval_df = self._evaluate_composite_pnl(pnl_xcat)
            else:
                pnl = self.single_return_pnls[pnl_xcat]
                eval_df = pnl.evaluate_pnls([pnl_xcat.split("/")[0]])
                eval_df.columns = [pnl_xcat]
            pnl_evals.append(eval_df)

        return pd.concat(pnl_evals, axis=1, ignore_index=False, sort=False)

    def _evaluate_composite_pnl(self, pnl_xcat: str) -> pd.DataFrame:
        """
        Evaluate the combined PnLs in a manner similar to NaivePnL's `evaluate_pnls()`.

        """
        stats = [
            "Return %",
            "St. Dev. %",
            "Sharpe Ratio",
            "Sortino Ratio",
            "Max 21-Day Draw %",
            "Max 6-Month Draw %",
            "Peak to Trough Draw %",
            "Top 5% Monthly PnL Share",
            "Traded Months",
        ]
        pnl_df = self.get_pnls([pnl_xcat])
        dfw = pnl_df.pivot(index="real_date", columns="xcat", values="value")
        df = pd.DataFrame(columns=dfw.columns, index=stats)

        df.iloc[0, :] = dfw.mean(axis=0) * 261
        df.iloc[1, :] = dfw.std(axis=0) * np.sqrt(261)
        df.iloc[2, :] = df.iloc[0, :] / df.iloc[1, :]
        dsd = dfw.apply(lambda x: np.sqrt(np.sum(x[x < 0] ** 2) / len(x))) * np.sqrt(
            261
        )
        df.iloc[3, :] = df.iloc[0, :] / dsd
        df.iloc[4, :] = dfw.rolling(21).sum().min()
        df.iloc[5, :] = dfw.rolling(6 * 21).sum().min()

        cum_pnl = dfw.cumsum()
        high_watermark = cum_pnl.cummax()
        drawdown = high_watermark - cum_pnl

        df.iloc[6, :] = -drawdown.max()

        monthly_pnl = dfw.resample("M").sum()
        total_pnl = monthly_pnl.sum(axis=0)
        top_5_percent_cutoff = int(np.ceil(len(monthly_pnl) * 0.05))
        top_months = pd.DataFrame(columns=monthly_pnl.columns)
        for column in monthly_pnl.columns:
            top_months[column] = (
                monthly_pnl[column]
                .nlargest(top_5_percent_cutoff)
                .reset_index(drop=True)
            )
        df.iloc[7, :] = top_months.sum() / total_pnl
        df.iloc[8, :] = dfw.resample("M").sum().count()
        return df

    def get_pnls(self, pnl_xcats: List[str] = None) -> pd.DataFrame:
        """
        Returns a DataFrame with PnLs.

        :param pnl_xcats: List of PnLs to return. If None, all PnLs are returned.
            Must be in the format 'xcat', or 'xcat/return_xcat'.
        """
        if self.pnls_df is None:
            raise ValueError("The PnLs have been added. Use add_pnl() first.")

        if pnl_xcats is None:
            return self.pnls_df

        else:
            for i, pnl_xcat in enumerate(pnl_xcats):
                pnl_xcats[i] = self._infer_return_by_xcat(pnl_xcat)

            return self.pnls_df[self.pnls_df["xcat"].isin(pnl_xcats)]

    def _normalize_weights(self, weights: dict) -> dict:
        """
        Normalize the weights to sum up to 1.
        """
        weights_sum = sum(weights.values())
        return {k: v / weights_sum for k, v in weights.items()}

    def _validate_pnl(self, pnl: NaivePnL, pnl_xcats: List[str]):
        """
        Validate the PnL and PnL categories.
        """
        if not isinstance(pnl, NaivePnL):
            raise ValueError("The pnl must be a NaivePnL object.")
        if not set(pnl_xcats).issubset(pnl.pnl_names):
            raise ValueError("The pnl_xcats must be in the NaivePnL object.")
        if not all(isinstance(x, str) for x in pnl_xcats):
            raise ValueError("All elements in the list must be strings.")
        return True

    def pnl_xcats(self):
        """
        Return all PnL categories.
        """
        return self.pnls_df["xcat"].unique().tolist()

    def return_xcats(self):
        """
        Return all return categories associated with PnLs.
        """
        return list(set(self.xcat_to_ret.values()))

    def _check_pnls_added(self, min_pnls: int = 1):
        """
        Check if at least `min_pnls` PnLs have been added.
        """
        if len(self.pnl_xcats()) < min_pnls:
            raise ValueError(
                f"At least {min_pnls} PnL must be added with add_pnl() first."
            )

    def _infer_return_by_xcat(self, pnl_xcat):
        """
        Infer the return category from the xcat if not provided.

        Throws an error is there are multiple return categories for the xcat.
        """
        if pnl_xcat in self.composite_pnl_xcats:
            return pnl_xcat

        if "/" not in pnl_xcat:
            if pnl_xcat not in self.xcat_to_ret:
                raise ValueError(f"{pnl_xcat} has not been added with add_pnl() yet.")
            if len(self.xcat_to_ret[pnl_xcat]) > 1:
                raise ValueError(
                    f"{pnl_xcat} corresponds to multiple return categories: {self.xcat_to_ret[pnl_xcat]}. "
                    "Must append return to xcat in the format 'xcat/return'."
                )
            else:
                return f"{pnl_xcat}/{list(self.xcat_to_ret[pnl_xcat])[0]}"
        else:
            if pnl_xcat not in self.pnl_xcats():
                raise ValueError(f"{pnl_xcat} has not been added with add_pnl() yet.")
            else:
                return pnl_xcat
            
    def _check_xcat_labels(self, pnl_xcats, xcat_labels):
        if isinstance(xcat_labels, dict):
            xcat_labels = {
                    self._infer_return_by_xcat(k): v for k, v in xcat_labels.items()
                }
        elif isinstance(xcat_labels, list):
            if len(pnl_xcats) != len(xcat_labels):
                raise ValueError(
                        "If using a list, the number of labels must match the number of PnLs."
                    )
            xcat_labels = dict(zip(pnl_xcats, xcat_labels))
        else:
            raise ValueError("xcat_labels must be a list or a dictionary.")
        return xcat_labels


if __name__ == "__main__":

    np.random.seed(0)

    cids = ["AUD", "CAD", "GBP", "NZD", "USD", "EUR"]
    xcats = ["EQXR_NSA", "FXXR", "GROWTH", "INFL", "DUXR"]

    cols_1 = ["earliest", "latest", "mean_add", "sd_mult"]
    df_cids = pd.DataFrame(index=cids, columns=cols_1)
    df_cids.loc["AUD", :] = ["2008-01-03", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD", :] = ["2010-01-03", "2020-11-30", 0, 1]
    df_cids.loc["GBP", :] = ["2012-01-03", "2020-11-30", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2002-01-03", "2020-09-30", -0.1, 2]
    df_cids.loc["USD"] = ["2015-01-03", "2020-12-31", 0.2, 2]
    df_cids.loc["EUR"] = ["2008-01-03", "2020-12-31", 0.1, 2]

    cols_2 = cols_1 + ["ar_coef", "back_coef"]

    df_xcats = pd.DataFrame(index=xcats, columns=cols_2)
    df_xcats.loc["EQXR_NSA"] = ["2000-01-03", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["FXXR"] = ["2000-01-01", "2020-10-30", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-03", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2001-01-01", "2020-10-30", 1, 2, 0.8, 0.5]
    df_xcats.loc["DUXR"] = ["2000-01-01", "2020-12-31", 0.1, 0.5, 0, 0.1]

    black = {"AUD": ["2006-01-01", "2015-12-31"], "GBP": ["2022-01-01", "2100-01-01"]}
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Instantiate a new instance to test the long-only functionality.
    # Benchmarks are used to calculate correlation against PnL series.
    pnl_eq = NaivePnL(
        dfd,
        ret="EQXR_NSA",
        sigs=["GROWTH"],
        cids=cids,
        start="2000-01-01",
        blacklist=black,
        # bms=["EUR_EQXR_NSA", "USD_EQXR_NSA"],
    )

    pnl_eq.make_pnl(
        sig="GROWTH",
        sig_op="zn_score_pan",
        sig_neg=False,
        sig_add=0.5,
        rebal_freq="monthly",
        vol_scale=5,
        rebal_slip=1,
        min_obs=250,
        thresh=2,
        pnl_name="PNL_EQ",
    )

    pnl_eq.make_long_pnl(vol_scale=10, label="LONG")

    pnl_fx = NaivePnL(
        dfd,
        ret="FXXR",
        sigs=["INFL"],
        cids=cids,
        start="2000-01-01",
        blacklist=black,
        bms=["EUR_EQXR_NSA", "USD_EQXR_NSA"],
    )

    pnl_fx.make_pnl(
        sig="INFL",
        sig_op="zn_score_pan",
        sig_neg=True,
        sig_add=0.5,
        rebal_freq="monthly",
        vol_scale=5,
        rebal_slip=1,
        min_obs=250,
        thresh=2,
        pnl_name="PNL_FX",
    )

    pnl_fx.make_long_pnl(vol_scale=10, label="LONG")

    # pnl_fx.make_long_pnl(vol_scale=10, label="Long")
    print(pnl_eq.pnl_names)

    mapnl = MultiPnL()

    mapnl.add_pnl(pnl_fx, ["PNL_FX", "LONG"])
    mapnl.add_pnl(pnl_eq, ["PNL_EQ", "LONG"])

    mapnl.combine_pnls(
        # ["PNL_FX", "PNL_EQ"],
        ["PNL_EQ", "PNL_FX"],
        # weights={"PNL_FX": 1, "LONG": 1},
        composite_pnl_xcat="EQ_FX_LONG",
    )

    mapnl.plot_pnls(["PNL_FX", "PNL_EQ"], xcat_labels=["z", "FX"], title="PnLs")
    # print(mapnl.get_pnls(["PNL_FX"]))
    # print(mapnl.evaluate_pnls(["PNL_EQ"]))



"""
Module with functionality for generating mock 
quantamental data for testing purposes.
"""

import numpy as np
import pandas as pd
from statsmodels.tsa.arima_process import ArmaProcess
from typing import List, Dict, Union, Optional
from collections import defaultdict
import datetime
import warnings
from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.management.utils import is_valid_iso_date, get_cid, get_xcat


def simulate_ar(nobs: int, mean: float = 0, sd_mult: float = 1, ar_coef: float = 0.75):
    """
    Create an auto-correlated data-series as numpy array.

    :param <int> nobs: number of observations.
    :param <float> mean: mean of values, default is zero.
    :param <float> sd_mult: standard deviation multipliers of values, default is 1.
        This affects non-zero means.
    :param <float> ar_coef: autoregression coefficient (between 0 and 1): default is 0.75.

    :return <np.ndarray>: autocorrelated data series.
    """

    # Define relative parameters for creating an AR process.
    ar_params = np.r_[1, -ar_coef]
    ar_proc = ArmaProcess(ar_params)  # define ARMA process
    ser = ar_proc.generate_sample(nobs)
    ser = ser + mean - np.mean(ser)
    return sd_mult * ser / np.std(ser)


def dataframe_generator(
    df_cids: pd.DataFrame, df_xcats: pd.DataFrame, cid: str, xcat: str
):
    """
    Adjacent method used to construct the quantamental DataFrame.

    :param <pd.DataFrame> df_cids:
    :param <pd.DataFrame> df_xcats:
    :param <str> cid: individual cross-section.
    :param <str> xcat: individual category.

    :return <Tuple[pd.DataFrame, pd.DatetimeIndex]>: Tuple containing
        the quantamental DataFrame and a DatetimeIndex of the business days.
    """
    qdf_cols = ["cid", "xcat", "real_date", "value"]

    sdate = pd.to_datetime(
        max(df_cids.loc[cid, "earliest"], df_xcats.loc[xcat, "earliest"])
    )
    edate = pd.to_datetime(
        min(df_cids.loc[cid, "latest"], df_xcats.loc[xcat, "latest"])
    )
    all_days = pd.date_range(sdate, edate)
    work_days = all_days[all_days.weekday < 5]

    df_add = pd.DataFrame(columns=qdf_cols)
    df_add["real_date"] = work_days
    df_add["cid"] = cid
    df_add["xcat"] = xcat

    return df_add, work_days


def make_qdf(df_cids: pd.DataFrame, df_xcats: pd.DataFrame, back_ar: float = 0):
    """
    Make quantamental DataFrame with basic columns: 'cid', 'xcat', 'real_date', 'value'.

    :param <pd.DataFrame> df_cids: DataFrame with parameters by cid. Row indices are
        cross-sections. Columns are:
        'earliest': string of earliest date (ISO) for which country values are available;
        'latest': string of latest date (ISO) for which country values are available;
        'mean_add': float of country-specific addition to any category's mean;
        'sd_mult': float of country-specific multiplier of an category's standard
            deviation.
    :param <pd.DataFrame> df_xcats: dataframe with parameters by xcat. Row indices are
        cross-sections. Columns are:
        'earliest': string of earliest date (ISO) for which category values are
        available;
        'latest': string of latest date (ISO) for which category values are available;
        'mean_add': float of category-specific addition;
        'sd_mult': float of country-specific multiplier of an category's standard
        deviation;
        'ar_coef': float between 0 and 1 denoting set auto-correlation of the category;
        'back_coef': float, coefficient with which communal (mean 0, SD 1) background
        factor is added to category values.
    :param <float> back_ar: float between 0 and 1 denoting set auto-correlation of the
        background factor. Default is zero.

    :return <pd.DataFrame>: basic quantamental DataFrame according to specifications.
    """
    df_list = []

    if any(df_xcats["back_coef"] != 0):
        sdate = min(min(df_cids.loc[:, "earliest"]), min(df_xcats.loc[:, "earliest"]))
        edate = max(max(df_cids.loc[:, "latest"]), max(df_xcats.loc[:, "latest"]))
        all_days = pd.date_range(sdate, edate)
        work_days = all_days[all_days.weekday < 5]
        ser = simulate_ar(len(work_days), mean=0, sd_mult=1, ar_coef=back_ar)
        df_back = pd.DataFrame(index=work_days, columns=["value"])
        df_back["value"] = ser

    for cid in df_cids.index:
        for xcat in df_xcats.index:
            df_add, work_days = dataframe_generator(
                df_cids=df_cids, df_xcats=df_xcats, cid=cid, xcat=xcat
            )

            ser_mean = df_cids.loc[cid, "mean_add"] + df_xcats.loc[xcat, "mean_add"]
            ser_sd = df_cids.loc[cid, "sd_mult"] * df_xcats.loc[xcat, "sd_mult"]
            ser_arc = df_xcats.loc[xcat, "ar_coef"]
            df_add["value"] = simulate_ar(
                len(work_days), mean=ser_mean, sd_mult=ser_sd, ar_coef=ser_arc
            )

            back_coef = df_xcats.loc[xcat, "back_coef"]
            # Add the influence of communal background series.
            if back_coef != 0:
                dates = df_add["real_date"]
                df_add["value"] = df_add["value"] + back_coef * df_back.loc[
                    dates, "value"
                ].reset_index(drop=True)

            df_list.append(df_add)

    return pd.concat(df_list).reset_index(drop=True)


def make_qdf_black(df_cids: pd.DataFrame, df_xcats: pd.DataFrame, blackout: dict):
    """
    Make quantamental DataFrame with basic columns: 'cid', 'xcat', 'real_date', 'value'.
    In this DataFrame the column, 'value', will consist of Binary Values denoting whether
    the cross-section is active for the corresponding dates.

    :param <pd.DataFrame> df_cids: dataframe with parameters by cid. Row indices are
        cross-sections. Columns are:
    'earliest': string of earliest date (ISO) for which country values are available;
    'latest': string of latest date (ISO) for which country values are available;
    'mean_add': float of country-specific addition to any category's mean;
    'sd_mult': float of country-specific multiplier of an category's standard deviation.
    :param <pd.DataFrame> df_xcats: dataframe with parameters by xcat. Row indices are
        cross-sections. Columns are:
    'earliest': string of earliest date (ISO) for which category values are available;
    'latest': string of latest date (ISO) for which category values are available;
    'mean_add': float of category-specific addition;
    'sd_mult': float of country-specific multiplier of an category's standard deviation;
    'ar_coef': float between 0 and 1 denoting set autocorrelation of the category;
    'back_coef': float, coefficient with which communal (mean 0, SD 1) background
        factor is added to categoy values.
    :param <dict> blackout: Dictionary defining the blackout periods for each cross-
        section. The expected form of the dictionary is:
        {'AUD': (Timestamp('2000-01-13 00:00:00'), Timestamp('2000-01-13 00:00:00')),
        'USD_1': (Timestamp('2000-01-03 00:00:00'), Timestamp('2000-01-05 00:00:00')),
        'USD_2': (Timestamp('2000-01-09 00:00:00'), Timestamp('2000-01-10 00:00:00')),
        'USD_3': (Timestamp('2000-01-12 00:00:00'), Timestamp('2000-01-12 00:00:00'))}
        The values of the dictionary are tuples consisting of the start & end-date of the
        respective blackout period. Each cross-section could have potentially more than
        one blackout period on a single category, and subsequently each key will be
        indexed to indicate the number of periods.

    :return <pd.DataFrame>: basic quantamental DataFrame according to specifications with
        binary values.
    """

    df_list = []

    conversion = lambda t: (pd.Timestamp(t[0]), pd.Timestamp(t[1]))
    dates_dict = defaultdict(list)
    for k, v in blackout.items():
        v = conversion(v)
        dates_dict[k[:3]].append(v)

    # At the moment the blackout period is being applied uniformally to each category:
    # each category will experience the same blackout periods.
    for cid in df_cids.index:
        for xcat in df_xcats.index:
            df_add, work_days = dataframe_generator(
                df_cids=df_cids, df_xcats=df_xcats, cid=cid, xcat=xcat
            )
            arr = np.repeat(0, df_add.shape[0])
            dates = df_add["real_date"].to_numpy()

            list_tuple = dates_dict[cid]
            for tup in list_tuple:
                start = tup[0]
                end = tup[1]

                condition_start = start.weekday() - 4
                condition_end = end.weekday() - 4

                # Will skip the associated blackout period because of the received
                # invalid date, if it is not within the respective data series' range,
                # but will continue to populate the dataframe according to the other keys
                # in the dictionary.
                # Naturally compare against the data-series' formal start & end date.
                if start < dates[0] or end > dates[-1]:
                    warnings.warn("Blackout period date not within data series range.")
                    break
                # If the date falls on a weekend, change to the following Monday.
                elif condition_start > 0:
                    while start.weekday() > 4:
                        start += datetime.timedelta(days=1)
                elif condition_end > 0:
                    while end.weekday() > 4:
                        end += datetime.timedelta(days=1)

                index_start = next(iter(np.where(dates == start)[0]))
                count = 0
                while start != tup[1]:
                    if start.weekday() < 5:
                        count += 1
                    start += datetime.timedelta(days=1)

                arr[index_start : (index_start + count + 1)] = 1

            df_add["value"] = arr

            df_list.append(df_add)

    return pd.concat(df_list).reset_index(drop=True)


def generate_lines(sig_len: int, style: str = "linear") -> Union[np.ndarray, List[str]]:
    """
    Returns a numpy array of a line with a given length.

    Parameters
    :param <int> sig_len: The number of elements in the returned array.
    :param <str> style: The style of the line. Default `'linear'`. Current choices are:
        `linear`, `decreasing-linear`, `sharp-hill`, `four-bit-sine`, `sine`, `cosine`,
        `sawtooth`. Adding `"inv"` or "inverted" to the style will return
        the inverted version of that line. For example, `'inv-sawtooth'` or `'inverted
        sawtooth'` will return the inverted sawtooth line. `'any'` will return a random
        line. `'all'` will return a list of all the available styles.

    :return <Union[np.ndarray, List[str]]>: A numpy array of the line. If `style` is
        `'all'`, then a list (of strings) of all the available styles is returned.

    NOTE: It is indeed request an `"inverted linear"` or `"inverted decreasing-linear"`
    line. They're just there for completeness and readability.
    """

    def _sawtooth(sig_len: int) -> np.ndarray:
        max_cycles = 4
        _tmp = sig_len // max_cycles
        up = np.tile(np.arange(-100, 100, 200 / _tmp), max_cycles * 2)
        return np.abs(up)[:sig_len]

    def _sharp_hill(sig_len: int) -> np.ndarray:
        return np.concatenate(
            [
                np.arange(1, sig_len // 4 + 1) * 100 / sig_len,
                np.arange(sig_len // 4, sig_len // 4 * 3 + 1)[::-1] * 100 / sig_len,
                np.arange(sig_len // 4 * 3, sig_len + 1) * 100 / sig_len,
            ]
        )

    def _four_bit_sine(sig_len: int) -> np.ndarray:
        return np.concatenate(
            [
                np.arange(1, sig_len // 4 + 1) * 100 / sig_len,
                np.arange(sig_len // 4, sig_len // 4 * 3 + 1)[::-1] * 100 / sig_len,
                np.arange(sig_len // 4 * 3, sig_len + 1) * 100 / sig_len,
            ]
        )

    def _sine(sig_len: int) -> np.ndarray:
        return np.sin(np.arange(1, sig_len + 1) * np.pi / (sig_len / 2)) * 50 + 50

    def _cosine(sig_len: int) -> np.ndarray:
        return np.cos(np.arange(1, sig_len + 1) * np.pi / (sig_len / 2)) * 50 + 50

    def _linear(sig_len: int) -> np.ndarray:
        return np.arange(1, sig_len + 1) * 100 / sig_len

    def _decreasing_linear(sig_len: int) -> np.ndarray:
        return _linear(sig_len)[::-1]

    if not isinstance(sig_len, int):
        raise TypeError("`sig_len` must be an integer.")
    elif sig_len < 1:
        raise ValueError("`sig_len` must be greater than 0.")

    if not isinstance(style, str):
        raise TypeError("`style` must be a string.")

    style: str = "-".join(style.strip().lower().split())

    lines: Dict[str, np.ndarray] = {
        "linear": _linear,
        "decreasing-linear": _decreasing_linear,
        "sharp-hill": _sharp_hill,
        "four-bit-sine": _four_bit_sine,
        "sine": _sine,
        "cosine": _cosine,
        "sawtooth": _sawtooth,
    }

    if "inv" in style:
        style = "-".join([s for s in style.split("-") if "inv" not in s])

    if style in lines:
        return lines[style](sig_len)[:sig_len]
    elif style == "any":
        r_choice: str = np.random.choice(list(lines.keys()))
        return lines[r_choice](sig_len)[:sig_len]
    elif style == "all":
        # return the list of choices
        opns: List[str] = list(lines.keys())
        inv_opns: List[str] = [f"inverted-{opn}" for opn in opns]
        return opns + inv_opns
    else:
        raise ValueError(f"Invalid style: {style}. Use one of: {list(lines.keys())}.")


def make_test_df(
    cids: Optional[List[str]] = ["AUD", "CAD", "GBP"],
    xcats: Optional[List[str]] = ["XR", "CRY"],
    tickers: Optional[List[str]] = None,
    metrics: List[str] = ["value"],
    start: str = "2010-01-01",
    end: str = "2020-12-31",
    style: str = "any",
) -> QuantamentalDataFrame:
    """
    Generates a test dataframe with pre-defined values.
    These values are meant to be used for testing purposes only.
    The functions generates a standard quantamental dataframe with
    where the value column is populated with pre-defined values.
    These values are simple lines, or waves that are easy to identify
    and differentiate in a plot.

    Parameters

    :param <List[str]> cids: A list of strings for cids.
    :param <List[str]> xcats: A list of strings for xcats.
    :param <List[str]> metrics: A list of strings for metrics.
    :param <str> start: An ISO-formatted date string.
    :param <str> end: An ISO-formatted date string.
    :param <str> style: A string that specifies the type of line to generate.
        Current choices are: 'linear', 'decreasing-linear', 'sharp-hill',
        'four-bit-sine', 'sine', 'cosine', 'sawtooth', 'any'. See
        `macrosynergy.management.simulate.simulate_quantamental_data.generate_lines()`.
    """
    ## Check the inputs
    if isinstance(cids, str):
        cids = [cids]
    if isinstance(xcats, str):
        xcats = [xcats]
    if isinstance(tickers, str):
        tickers = [tickers]
    if isinstance(metrics, str):
        metrics = [metrics]
    if not isinstance(metrics, list):
        raise TypeError("`metrics` must be a list of strings.")

    if "all" in metrics:
        metrics = ["value", "grading", "eop_lag", "mop_lag"]

    if (cids is None) != (xcats is None):
        raise ValueError("Please provide both `cids` and `xcats` or neither.")

    if tickers is None or len(tickers) == 0:
        if cids is None:
            raise ValueError("Please provide a list of tickers or `cids` & `xcats`.")

    for varx, namex in zip(
        [cids, xcats, metrics, tickers], ["cids", "xcats", "metrics", "tickers"]
    ):
        if varx is not None:
            if not isinstance(varx, list):
                raise TypeError(f"`{namex}` must be a list.")
            if len(varx) == 0:
                raise ValueError(f"`{namex}` cannot be empty.")
            if not all(isinstance(x, str) for x in varx):
                raise TypeError(f"All elements in `{namex}` must be strings.")

    for varx, namex in zip([start, end], ["start", "end"]):
        if not is_valid_iso_date(varx):
            raise ValueError(f"`{namex}` must be a valid ISO date string.")

    ## Generate the dataframe

    dates: pd.DatetimeIndex = pd.bdate_range(start, end)
    all_tickers: List[str] = tickers if tickers is not None else []
    if cids is not None:
        all_tickers += [f"{cid}_{xcat}" for cid in cids for xcat in xcats]
    all_tickers = sorted(set(all_tickers))
    df_list: List[pd.DataFrame] = []

    for ticker in all_tickers:
        df_add: pd.DataFrame = pd.DataFrame(
            index=dates, columns=["real_date", "cid", "xcat", *metrics]
        )
        df_add["cid"] = get_cid(ticker)
        df_add["xcat"] = get_xcat(ticker)
        df_add["real_date"] = dates
        for metric in metrics:
            df_add[metric] = generate_lines(len(dates), style=style)
        df_list.append(df_add)

    return pd.concat(df_list).reset_index(drop=True)


if __name__ == "__main__":
    ser_ar = simulate_ar(100, mean=0, sd_mult=1, ar_coef=0.75)

    cids = ["AUD", "CAD", "GBP"]
    xcats = ["XR", "CRY"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD",] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD",] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP",] = ["2011-01-01", "2020-11-30", -0.2, 0.5]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR",] = ["2010-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcats.loc["CRY",] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
